from enum import auto
from mstrio.utils.enum_helper import AutoName, AutoCapitalizedName


class DBType(AutoName):
    """Enumeration representing database type."""

    RESERVED = auto()
    ACCESS = auto()
    ALTIBASE = auto()
    AMAZON_ATHENA = auto()
    AMAZON_AURORA = auto()
    AMAZON_DOCUMENT_DB = auto()
    AMAZON_DYNAMO_DB = auto()
    AMAZON_REDSHIFT = auto()
    ARCADIA_PLATFORM = auto()
    ASTER = auto()
    AZURE_COSMOS = auto()
    BIG_DATA_ENGINE = auto()
    CASSANDRA = auto()
    CIRRO = auto()
    CLOUD_ELEMENT = auto()
    CLOUD_GATEWAY = auto()
    CLOUD_GATEWAY_AWS_S3 = auto()
    CLOUD_GATEWAY_AZURE_ADLS_2 = auto()
    CLOUD_GATEWAY_GOOGLE_CLOUD_STORAGE = auto()
    COMPOSITE = auto()
    CONCUR = auto()
    CONNECTION_CLOUD = auto()
    DATABRICKS = auto()
    DATA_DIRECT_CLOUD = auto()
    DATALLEGRO = auto()
    DB2 = auto()
    DENODO = auto()
    DREMIO = auto()
    DRILL = auto()
    DROPBOX = auto()
    DRUID = auto()
    ELASTICSEARCH = auto()
    ELOQUA = auto()
    ENTERPRISE_DB = auto()
    ESS_BASE = auto()
    EXA_SOLUTION = auto()
    EXCEL = auto()
    FACEBOOK = auto()
    FINANCIALFORCE = auto()
    GBASE_8A = auto()
    GENERIC = auto()
    GENERIC_DATA_CONNECTOR = auto()
    GITHUB = auto()
    GOOGLE_ANALYTICS = auto()
    GOOGLE_BIG_QUERY = auto()
    GOOGLE_BIG_QUERY_FF_SQL = auto()
    GOOGLE_CLOUD_SPANNER = auto()
    GOOGLE_DRIVE = auto()
    HIVE = auto()
    HIVE_THRIFT = auto()
    HUBSPOT = auto()
    IBM_IPS = auto()
    IMPALA = auto()
    INFORMATICA = auto()
    INFORMIX = auto()
    JIRA = auto()
    KAFKA = auto()
    KOGNITIOWX2 = auto()
    KYVOS_MDX = auto()
    MAPD = auto()
    MARKETO = auto()
    MARK_LOGIC = auto()
    MEM_SQL = auto()
    METAMATRIX = auto()
    MICROSOFT_AS = auto()
    MICROSOFT_DYNAMICS_CRM = auto()
    MICROSOFT_DYNAMICS_ERP = auto()
    MICROSOFT_DYNAMICS_365 = auto()
    MONGO_BI = auto()
    MONGO_DB = auto()
    MY_SQL = auto()
    NEO4J = auto()
    NEOVIEW = auto()
    NETEZZA = auto()
    ODATA = auto()
    OPEN_ACCESS = auto()
    ORACLE = auto()
    ORACLE_CX_SALES = auto()
    ORACLE_CX_SERVICE = auto()
    ORACLE_ELOQUA = auto()
    PALANTIR_FOUNDRY = auto()
    PAR_ACCEL = auto()
    PAR_STREAM = auto()
    PAYPAL = auto()
    PHOENIX = auto()
    PIG = auto()
    PIVOTAL_HAWQ = auto()
    POSTGRE_SQL = auto()
    PRESTO = auto()
    PYTHON = auto()
    RED_BRICK = auto()
    SALESFORCE = auto()
    SAND = auto()
    SAP = auto()
    SAP_BW4_HANA = auto()
    SAP_BW_ODATA = auto()
    SAP_ECC_ODATA = auto()
    SAP_HANA = auto()
    SAP_HANA_MDX = auto()
    SAP_S4_HANA = auto()
    SEARCH_ENGINE = auto()
    SERVICEMAX = auto()
    SERVICENOW = auto()
    SHOPIFY = auto()
    SNOW_FLAKE = auto()
    SPARK_CONFIG = auto()
    SPARK_SQL = auto()
    SPLUNK = auto()
    SQL_SERVER = auto()
    SQUARE = auto()
    STARBURST = auto()
    SUGAR_CRM = auto()
    SYBASE = auto()
    SYBASE_IQ = auto()
    SYBASE_SQL_ANY = auto()
    TANDEM = auto()
    TEAMCITY = auto()
    TERADATA = auto()
    TM1 = auto()
    TRINO = auto()
    TWITTER = auto()
    UNKNOWN = auto()
    URL_AUTH = auto()
    VECTORWISE = auto()
    VERTICA = auto()
    XQUERY = auto()
    YELLOWBRICK = auto()


class GatewayType(AutoCapitalizedName):
    """Enumeration representing gateway type."""

    CLOUD = auto()
    COMMUNITY = auto()
    NATIVE = auto()
    RELATIONAL = auto()



"""List names and IDs of empty user groups."""

from typing import List

from mstrio.connection import Connection, get_connection
from mstrio.users_and_groups import list_user_groups, UserGroup


def list_empty_user_groups(connection: "Connection") -> List["UserGroup"]:
    """List empty user groups.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
    """
    all_user_groups = list_user_groups(connection=connection)
    return [
        user_group_ for user_group_ in all_user_groups if not user_group_.list_members()
    ]


# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, 'MicroStrategy Tutorial')

# get empty user groups
empty_user_groups = list_empty_user_groups(conn)

# print empty user groups
print('Empty user groups:', flush=True)
for ug in empty_user_groups:
    print(f"{ug.name} ({ug.id}) ", flush=True)



from typing import List

from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg='Error getting bookmarks from shortcut {shortcut_id}')
def get_bookmarks_from_shortcut(connection, shortcut_id, error_msg=None):
    """Get a Bookmark list from a Shortcut Object.

    Args:
        connection: MicroStrategy REST API connection object
        shortcut_id (string): Shortcut  ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}/api/shortcuts/{shortcut_id}/bookmarks"
    return connection.get(url=url)


@ErrorHandler(err_msg='Error updating information for shortcut {shortcut_id}')
def update_information_for_shortcut(connection, shortcut_id, body, error_msg=None):
    """Update info for a shortcut.

    Args:
        connection: MicroStrategy REST API connection object
        shortcut_id (string): The ID of the document shortcut to execute
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = connection.base_url + f"​/api/shortcuts/{shortcut_id}"
    return connection.patch(url=url, json=body)


@ErrorHandler(err_msg='Error adding a new bookmark.')
def refresh_document_instance(connection, error_msg=None):
    """Add a new bookmark into current shortcut object.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = connection.base_url + "​/api​/bookmarks"
    return connection.put(url=url)


@ErrorHandler(err_msg='Error deleting bookmarks.')
def delete_bookmarks(connection, shortcut_id: str, bookmark_ids: List, error_msg=None):
    """Bulk deletion of bookmarks.

    Args:
        connection: MicroStrategy REST API connection object
        shortcut_id: Shortcut ID
        bookmark_ids: IDs of bookmarks to be deleted
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    body = {"shortcutId": shortcut_id, "bookmarkIds": bookmark_ids}
    url = f'{connection.base_url}/api/bookmarks'
    return connection.delete(url=url, json=body)


@ErrorHandler(err_msg='Error deleting bookmark with ID {bookmark_id}')
def delete_single_bookmark(
    connection, shortcut_id: str, bookmark_id: str, error_msg=None
):
    """Delete a bookmark.

    Args:
        connection: MicroStrategy REST API connection object
        shortcut_id: Shortcut ID
        bookmark_id (string): Bookmark ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    body = {"shortcutId": shortcut_id}
    url = f'{connection.base_url}/api/bookmarks/{bookmark_id}'
    return connection.delete(url=url, json=body)


@ErrorHandler(err_msg='Error updating bookmark {bookmark_id}')
def update_bookmark(connection, bookmark_id, body, error_msg=None):
    """Update a bookmark.

    Args:
        connection: MicroStrategy REST API connection object
        bookmark_id (string): Bookmark ID
        body: JSON-formatted information used to format the document
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    url = f"{connection.base_url}/api/bookmarks/{bookmark_id}"
    return connection.put(url=url, json=body)


@ErrorHandler(err_msg='Error adding bookmark {bookmark_name}')
def add_bookmark(
    connection, bookmark_name, instance_id, shortcut_id, error_msg=None
):  # shortcut needs to be created to add a bookmark!
    """Update a bookmark.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    body = {"name": bookmark_name, "instanceId": instance_id}
    url = f'{connection.base_url}/api/bookmarks'
    return connection.post(url=url, json=body)


@ErrorHandler(err_msg='Error getting shortcuts for document {document_id}')
def get_document_shortcut(connection, document_id, instance_id, error_msg=None):
    """Retrieve a published shortcut from the document definition.

    Args:
        connection: MicroStrategy REST API connection object
        document_id (string): Document  ID
        instance_id (string): Instance  ID
        error_msg (string, optional): Custom Error Message for Error Handling


    Returns:
        Complete HTTP response object.
    """
    endpoint_url = f'/api/documents/{document_id}/instances/{instance_id}/shortcut'
    return connection.get(connection.base_url + endpoint_url)



import logging
from typing import Optional, TYPE_CHECKING

from mstrio import config
from mstrio.api import folders, objects
from mstrio.object_management import PredefinedFolders
from mstrio.types import ObjectTypes
from mstrio.users_and_groups import User
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, MoveMixin
from mstrio.utils.helper import (
    fetch_objects_async,
    get_default_args_from_func,
    get_valid_project_id,
)

if TYPE_CHECKING:
    from mstrio.connection import Connection

logger = logging.getLogger(__name__)


def list_folders(
    connection: "Connection",
    project_id: Optional[str] = None,
    project_name: Optional[str] = None,
    to_dictionary: bool = False,
    limit: Optional[int] = None,
    **filters,
) -> list["Folder"] | list[dict]:
    """Get a list of folders - either all folders in a specific project or all
    folders that are outside of projects, called configuration-level folders.
    The list of configuration-level folders includes folders such as users, user
    groups, databases, etc. which are not project-specific. If you pass
    a `project_id` or `project_name`, you get folders in that project;
    if not, then you get configuration-level folders.

    Specify either `project_id` or `project_name`.
    When `project_id` is provided (not `None`), `project_name` is omitted.

    Note:
        Id of project is not taken directly from `Connection` object, so you
        have to specify it explicitly.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        project_id (string, optional): project ID
        project_name (string, optional): project name
        to_dictionary (bool, optional): If True returns dicts, by default
            (False) returns objects.
        limit (int): limit the number of elements returned. If `None` (default),
            all objects are returned.
        **filters: Available filter parameters: ['name', 'id', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'acg',
             'owner', 'hidden',
          'ext_type']

    Returns:
        list of `Folder` objects or list of dictionaries
    """
    # Project is validated only if project was specified in arguments -
    # otherwise fetch is performed from a non-project area.
    if project_id or project_name:
        project_id = get_valid_project_id(
            connection=connection,
            project_id=project_id,
            project_name=project_name,
        )
    objects = fetch_objects_async(
        connection,
        folders.list_folders,
        folders.list_folders_async,
        limit=limit,
        chunk_size=1000,
        project_id=project_id,
        filters=filters,
    )

    if to_dictionary:
        return objects
    else:
        from mstrio.utils.object_mapping import map_objects_list

        return map_objects_list(connection, objects)


def get_my_personal_objects_contents(
    connection: "Connection",
    project_id: Optional[str] = None,
    project_name: Optional[str] = None,
    to_dictionary: bool = False,
) -> list:
    """Get contents of `My Personal Objects` folder in a specific project.

    Specify either `project_id` or `project_name`.
    When `project_id` is provided (not `None`), `project_name` is omitted.

    Note:
        When `project_id` is `None` and `project_name` is `None`,
        then its value is overwritten by `project_id` from `connection` object.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        project_id (string, optional): project ID
        project_name (string, optional): project name
        to_dictionary (bool, optional): If True returns dicts, by default
            (False) returns objects.

    Returns:
        list of objects or list of dictionaries
    """
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=False if project_name else True,
    )

    objects = folders.get_my_personal_objects_contents(connection, project_id).json()
    if to_dictionary:
        return objects
    else:
        from mstrio.utils.object_mapping import map_objects_list

        return map_objects_list(connection, objects)


def get_predefined_folder_contents(
    connection: "Connection",
    folder_type: PredefinedFolders,
    project_id: Optional[str] = None,
    project_name: Optional[str] = None,
    to_dictionary: bool = False,
    limit: Optional[int] = None,
    **filters,
) -> list:
    """Get contents of a pre-defined MicroStrategy folder in a specific project.
    Available values for `folder_type` are stored in enum `PredefinedFolders`.

    Specify either `project_id` or `project_name`.
    When `project_id` is provided (not `None`), `project_name` is omitted.

    Note:
        When `project_id` is `None` and `project_name` is `None`,
        then its value is overwritten by `project_id` from `connection` object.

    Note:
        When `project_id` is `None`, then its value is overwritten by
        `project_id` from `connection` object.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`
        folder_type (enum): pre-defined folder type. Available values are
            stored in enum `PredefinedFolders`.
        project_id (string, optional): project ID
        project_name (string, optional): project name
        to_dictionary (bool, optional): If True returns dicts, by default
            (False) returns objects.
        limit (int): limit the number of elements returned. If `None` (default),
            all objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'acg']

    Returns:
        list of objects or list of dictionaries
    """

    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=False if project_name else True,
    )

    objects = fetch_objects_async(
        connection,
        folders.get_predefined_folder_contents,
        folders.get_predefined_folder_contents_async,
        limit=limit,
        chunk_size=1000,
        folder_type=folder_type.value,
        project_id=project_id,
        filters=filters,
    )

    if to_dictionary:
        return objects
    else:
        from mstrio.utils.object_mapping import map_objects_list

        return map_objects_list(connection, objects)


class Folder(Entity, CopyMixin, MoveMixin, DeleteMixin):
    """Object representation of MicroStrategy Folder object.

    Attributes:
        connection: MSTR Connection object
        id: ID of the folder
        name: name of the folder
        description: description of the folder
        type: the type of the folder (based on EnumDSSXMLObjectTypes)
        subtype: the subtype of the folder (based on EnumDSSXMLObjectSubTypes)
        date_created: the date/time at which the folder was first saved into the
            metadata
        date_modified: the date/time at which the folder was last saved into the
            metadata
        version: the version number this folder is currently carrying
        acl: an array of access control entry objects
        owner: User object, the owner of the object
        acg: access rights, EnumDSSXMLAccessRightFlags
        ext_type: object extended type, EnumDSSExtendedType
        ancestors: list of ancestor folders
        abbreviation: folder's abbreviation
        hidden: specifies whether the folder is hidden
        icon_path: folder icon path
        view_media: view media settings
        certified_info: CertifiedInfo object, certification status, time of
            certification, and information about the certifier (currently only
            for document and report)
        contents: contents of folder
    """

    _FROM_DICT_MAP = {**Entity._FROM_DICT_MAP, 'owner': User.from_dict}

    _API_PATCH: dict = {
        **Entity._API_PATCH,
        ('folder_id'): (objects.update_object, 'partial_put'),
    }

    _OBJECT_TYPE = ObjectTypes.FOLDER
    _SIZE_LIMIT = 10000000  # this sets desired chunk size in bytes

    def __init__(self, connection: "Connection", id: str, name: Optional[str] = None):
        """Initialize folder object by its identifier.

        Note:
            Parameter `name` is not used when fetching. `id` is always used to
            uniquely identify folder.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            id (str): Identifier of a pre-existing folder containing
                the required data.
            name (str): name of folder.
        """
        super().__init__(connection, id, name=name)

    @classmethod
    def create(
        cls,
        connection: "Connection",
        name: str,
        parent: str,
        description: Optional[str] = None,
    ) -> "Folder":
        """Create a new folder in a folder selected within connection object
        by providing its name, id of parent folder and optionally description.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            name (str): name of a new folder.
            parent (str): id of a parent folder in which new folder will be
                created.
            description (str, optional): optional description of a new folder.

        Returns:
            newly created folder
        """
        connection._validate_project_selected()
        response = folders.create_folder(connection, name, parent, description).json()
        if config.verbose:
            logger.info(
                f"Successfully created folder named: '{response.get('name')}' "
                f"with ID: '{response.get('id')}'"
            )
        return cls.from_dict(source=response, connection=connection)

    def alter(
        self, name: Optional[str] = None, description: Optional[str] = None
    ) -> None:
        """Alter the folder properties.

        Args:
            name: folder name
            description: folder description
        """
        func = self.alter
        default_dict = get_default_args_from_func(func)
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    def get_contents(self, to_dictionary: bool = False, **filters) -> list:
        """Get contents of folder. It can contains other folders or different
        kinds of objects.

        Args:
            to_dictionary (bool, optional): If True returns dicts, by default
                (False) returns objects.
            **filters: Available filter parameters: ['id', 'name',
                'description', 'date_created', 'date_modified', 'acg']

        Returns:
            Contents as Python objects (when `to_dictionary` is `False` (default
            value)) or contents as dictionaries otherwise.
        """
        objects = fetch_objects_async(
            self.connection,
            folders.get_folder_contents,
            folders.get_folder_contents_async,
            limit=None,
            chunk_size=1000,
            id=self.id,
            filters=filters,
        )

        if to_dictionary:
            return objects
        else:
            from mstrio.utils.object_mapping import map_objects_list

            return map_objects_list(self.connection, objects)



# flake8: noqa
from .enums import *
from .fact_expression import FactExpression
from .expression import Expression, Token, list_functions
from .expression_nodes import (
    AttributeFormPredicate,
    BandingCountPredicate,
    BandingDistinctPredicate,
    BandingPointsPredicate,
    BandingSizePredicate,
    ColumnReference,
    Constant,
    CustomExpressionPredicate,
    DynamicDateTime,
    ElementListPredicate,
    ExpressionFormShortcut,
    ExpressionRelationship,
    FilterQualificationPredicate,
    JointElementListPredicate,
    MetricPredicate,
    Operator,
    ObjectReference,
    PromptPredicate,
    ReportQualificationPredicate,
    SetFromRelationshipPredicate,
)
from .parameters import (
    AttributeElement,
    ConstantArrayParameter,
    ConstantParameter,
    DynamicDateTimeParameter,
    ExpressionParameter,
    FunctionProperty,
    ObjectReferenceParameter,
    PromptParameter,
    Variant,
    VariantType,
)
from .dynamic_date_time import (
    AdjustmentMonthlyByDay,
    AdjustmentMonthlyByDayOfWeek,
    AdjustmentMonthlyByReverseCount,
    AdjustmentNone,
    AdjustmentQuarterlyByDay,
    AdjustmentQuarterlyByDayOfWeek,
    AdjustmentQuarterlyByReverseCount,
    AdjustmentWeeklyByDayOfWeek,
    AdjustmentYearlyByDate,
    AdjustmentYearlyByDayOfWeek,
    DayOfWeek,
    DateMode,
    DynamicVersatileDate,
    DynamicDateTimeStructure,
    DynamicDateTimeType,
    HourMode,
    MinuteAndSecondMode,
    StaticVersatileDate,
    VersatileTime,
)



import logging
from typing import TYPE_CHECKING

from mstrio.types import ObjectTypes
from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from requests_futures.sessions import FuturesSession

logger = logging.getLogger(__name__)


@ErrorHandler(err_msg='Error getting information for the object with ID {id}')
def get_object_info(
    connection,
    id,
    object_type,
    project_id=None,
    error_msg=None,
    whitelist=[('ERR001 ', 500)],
):
    """Get information for a specific object in a specific project; if you do
    not specify a project ID, you get information for the object just in the
    non-project area.

    You identify the object with the object ID and object type. You specify
    the object type as a query parameter; possible values for object type are
    provided in EnumDSSXMLObjectTypes.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        object_type (int): One of EnumDSSXMLObjectTypes. Ex. 34 (User or
        UserGroup), 44 (Security Role), 32 (Project), 8 (Folder), 36 (type of
        I-Server configuration), 58 (Security Filter)
        project_id(str): ID of a project in which the object is located.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    if object_type == ObjectTypes.PROJECT.value:
        headers = {'X-MSTR-ProjectID': None}
    elif project_id:
        headers = {'X-MSTR-ProjectID': project_id}
    else:
        headers = {'X-MSTR-ProjectID': connection.project_id}

    if not project_id and not connection.project_id:
        logger.info(
            'Project was not selected. Search is performed for the non-project area'
        )

    return connection.get(
        url=f'{connection.base_url}/api/objects/{id}',
        headers=headers,
        params={'type': object_type},
    )


def get_object_info_async(
    futures_session: "FuturesSession", connection, id, object_type, project_id=None
):
    """Get information for a specific object in a specific project; if you do
    not specify a project ID, you get information for the object just in the
    non-project area.

    You identify the object with the object ID and object type. You specify
    the object type as a query parameter; possible values for object type are
    provided in EnumDSSXMLObjectTypes.

    Args:
        futures_session (FuturesSession): Future Session object to call
            MicroStrategy REST Server asynchronously
        connection (Connection): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        object_type (int): One of EnumDSSXMLObjectTypes. Ex. 34 (User or
        UserGroup), 44 (Security Role), 32 (Project), 8 (Folder), 36 (type of
        I-Server configuration), 58 (Security Filter)
        project_id(str): ID of a project in which the object is located.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    if object_type == ObjectTypes.PROJECT.value:
        headers = {'X-MSTR-ProjectID': None}
    elif project_id:
        headers = {'X-MSTR-ProjectID': project_id}
    else:
        headers = {'X-MSTR-ProjectID': connection.project_id}

    if not project_id and not connection.project_id:
        logger.info(
            'Project was not selected. Search is performed for the non-project area'
        )

    return futures_session.get(
        url=f'{connection.base_url}/api/objects/{id}',
        headers=headers,
        params={'type': object_type},
    )


@ErrorHandler(err_msg='Error deleting object with ID {id}')
def delete_object(connection, id, object_type, project_id=None, error_msg=None):
    """Get information for a specific object in a specific project; if you do
    not specify a project ID, you get information for the object in all
    projects.

    You identify the object with the object ID and object type. You specify
    the object type as a query parameter; possible values for object type are
    provided in EnumDSSXMLObjectTypes.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        object_type (int): One of EnumDSSXMLObjectTypes. Ex. 34 (User or
        UserGroup), 44 (Security Role), 32 (Project), 8 (Folder), 36 (type of
        I-Server configuration)
        project_id(str): ID of a project in which the object is located.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """

    if object_type == 32:
        headers = {'X-MSTR-ProjectID': str(id)}
    elif project_id:
        headers = {'X-MSTR-ProjectID': project_id}
    else:
        headers = {'X-MSTR-ProjectID': connection.project_id}

    return connection.delete(
        url=f'{connection.base_url}/api/objects/{id}',
        headers=headers,
        params={'type': object_type},
    )


@ErrorHandler(err_msg='Error updating object with ID {id}')
def update_object(
    connection, id, body, object_type, project_id=None, error_msg=None, verbose=True
):
    """Get information for a specific object in a specific project; if you do
    not specify a project ID, you get information for the object in all
    projects.

    You identify the object with the object ID and object type. You specify
    the object type as a query parameter; possible values for object type are
    provided in EnumDSSXMLObjectTypes.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        body: (object): body of the response
        object_type (int): One of EnumDSSXMLObjectTypes. Ex. 34 (User or
        UserGroup), 44 (Security Role), 32 (Project), 8 (Folder), 36 (type of
        I-Server configuration)
        project_id(str): ID of a project in which the object is located.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    if object_type == 32:
        headers = {'X-MSTR-ProjectID': str(id)}
    elif project_id:
        headers = {'X-MSTR-ProjectID': project_id}
    else:
        headers = {'X-MSTR-ProjectID': connection.project_id}

    return connection.put(
        url=f'{connection.base_url}/api/objects/{id}',
        headers=headers,
        params={'type': object_type},
        json=body,
    )


@ErrorHandler(err_msg='Error creating a copy of object with ID {id}')
def copy_object(
    connection, id, name, folder_id, object_type, project_id=None, error_msg=None
):
    """Create a copy of a specific object.

    You identify the object with the object ID and object type. You obtain the
    authorization token needed to execute the request using POST /auth/login;
    you obtain the project ID using GET /projects. You pass the authorization
    token and the project ID in the request header. You specify the object ID in
    the path of the request and object type as a query parameter; possible
    values for object type are provided in EnumDSSXMLObjectTypes. You specify
    the name and location (folder ID) of the new object in the body of the
    request. If you do not specify a new name, a default name is generated, such
    as 'Old Name (1)'. If you do not specify a folder ID, the object is saved in
    the same folder as the source object.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        object_type (int): One of EnumDSSXMLObjectTypes. Ex. 34 (User or
        UserGroup), 44 (Security Role), 32 (Project), 8 (Folder), 36 (type of
        I-Server configuration)
        project_id(str): ID of a project in which the object is located.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    if object_type == 32:
        headers = {'X-MSTR-ProjectID': str(id)}
    elif project_id:
        headers = {'X-MSTR-ProjectID': project_id}
    elif connection.project_id:
        headers = {'X-MSTR-ProjectID': connection.project_id}
    else:
        raise ValueError("Project needs to be specified.")

    body = {"name": name, "folderId": folder_id}
    return connection.post(
        url=f'{connection.base_url}/api/objects/{id}/copy',
        headers=headers,
        params={'type': object_type},
        json=body,
    )


@ErrorHandler(err_msg='Error getting VLDB settings for object with ID {id}')
def get_vldb_settings(connection, id, object_type, project_id=None, error_msg=None):
    """Get vldb settings for an object.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        object_type (int): DssXmlTypeReportDefinition(3) for Dataset and
            DssXmlTypeDocumentDefinition(55) for document/dossier
        project_id(str): ID of a project in which the object is located.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    headers = {}
    if project_id:
        headers = {'X-MSTR-ProjectID': project_id}
    else:
        connection._validate_project_selected()
        headers = {'X-MSTR-ProjectID': connection.project_id}

    return connection.get(
        url=f"{connection.base_url}/api/objects/{id}/vldb/propertySets",
        params={'type': object_type},
        headers=headers,
    )


@ErrorHandler(
    err_msg='Error resetting all custom vldb settings for object with ID {id}'
)
def delete_vldb_settings(connection, id, object_type, project_id=None, error_msg=None):
    """Delete all customized vldb settings in one object, this operation will
    reset all vldb settings to default.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        object_type (int): DssXmlTypeReportDefinition(3) for Dataset and
            DssXmlTypeDocumentDefinition(55) for document/dossier
        project_id(str): ID of a project in which the object is located.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    headers = {}
    if project_id:
        headers = {'X-MSTR-ProjectID': project_id}
    else:
        connection._validate_project_selected()
        headers = {'X-MSTR-ProjectID': connection.project_id}

    return connection.delete(
        url=f"{connection.base_url}/api/objects/{id}/vldb/propertySets",
        params={'type': object_type},
        headers=headers,
    )


@ErrorHandler(
    err_msg='Error resetting all custom vldb settings for object with ID {id}'
)
def set_vldb_settings(
    connection, id, object_type, name, body, project_id=None, error_msg=None
):
    """Set vldb settings for one property set in one object.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        object_type (int): DssXmlTypeReportDefinition(3) for Dataset and
            DssXmlTypeDocumentDefinition(55) for document/dossier
        name: property set name
        body: [{"name": "string",
                "value": {}}]
        project_id(str): ID of a project in which the object is located.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object returned by the MicroStrategy REST server
    """
    headers = {}
    if project_id:
        headers = {'X-MSTR-ProjectID': project_id}
    else:
        connection._validate_project_selected()
        headers = {'X-MSTR-ProjectID': connection.project_id}

    return connection.put(
        url=f"{connection.base_url}/api/objects/{id}/vldb/propertySets/{name}",
        params={'type': object_type},
        headers=headers,
        json=body,
    )


@ErrorHandler(err_msg='Error getting objects.')
def create_search_objects_instance(
    connection,
    name=None,
    pattern=4,
    domain=2,
    root=None,
    object_type=None,
    error_msg=None,
):
    """Create a search instance.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        name: expression used with the pattern to do the search
        pattern: specifies the nature of the search. Possible values are defined
            in the EnumDSSXMLSearchTypes javadoc
        domain: search domain. specifies the domain/scope of the search.
            Possible values are defined in the EnumDSSXMLSearchDomain javadoc
        root: folder ID of the root in which the search is done
        object_type: specifies the type of objects to be searched. Possible
            values are defined in the EnumDSSObjectType javadoc
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response returned by the MicroStrategy REST server
    """
    connection._validate_project_selected()
    return connection.post(
        url=f"{connection.base_url}/api/objects",
        headers={'X-MSTR-ProjectID': connection.project_id},
        params={
            'name': name,
            'pattern': pattern,
            'domain': domain,
            'root': root,
            'type': object_type,
        },
    )


@ErrorHandler(err_msg='Error getting objects using search with ID {search_id}')
def get_objects(
    connection, search_id, offset=0, limit=-1, get_tree=False, error_msg=None
):
    """Get list of objects from metadata.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        search_id: ID for the results of a previous search stored in I-Server
            memory
        offset: starting point within the collection of returned results. Used
            to control paging behavior.
        limit: maximum number of items returned for a single request. Used to
            control paging behavior
        get_tree: specifies that the search results should be displayed in
            a tree structure instead of a list. The ancestors of the searched
            objects are the nodes and the searched objects are the leaves of
            the tree.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response returned by the MicroStrategy REST server
    """
    connection._validate_project_selected
    return connection.get(
        url=f"{connection.base_url}/api/objects",
        headers={'X-MSTR-ProjectID': connection.project_id},
        params={
            'searchId': search_id,
            'offset': offset,
            'limit': limit,
            'getTree': get_tree,
        },
    )


def get_objects_async(
    future_session: "FuturesSession",
    connection,
    search_id,
    offset=0,
    limit=-1,
    get_tree=False,
):
    """Get list of objects from metadata asynchronously.

    Args:
        future_session(object): Future Session object to call MicroStrategy REST
            Server asynchronously
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        search_id: ID for the results of a previous search stored in I-Server
            memory
        offset: starting point within the collection of returned results. Used
            to control paging behavior.
        limit: maximum number of items returned for a single request. Used to
            control paging behavior.
        get_tree: specifies that the search results should be displayed in
            a tree structure instead of a list. The ancestors of the searched
            objects are the nodes and the searched objects are the leaves of
            the tree.

    Returns:
        HTTP response returned by the MicroStrategy REST server
    """
    connection._validate_project_selected()
    url = connection.base_url + '/api/objects'
    headers = {'X-MSTR-ProjectID': connection.project_id}
    params = {
        'searchId': search_id,
        'offset': offset,
        'limit': limit,
        'getTree': get_tree,
    }
    future = future_session.get(url=url, headers=headers, params=params)
    return future


@ErrorHandler(err_msg='Error certifying object with ID {id}')
def toggle_certification(connection, id, object_type=3, certify=True):
    """Certify/Uncertify a multi-table dataset.

    Args:
        connection (object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Identifier of a pre-existing dataset. Used when
            certifying a pre-existing dataset.
        object_type (int, optional): Type of object to certify as integer;
            defaults to 3 (dataset)
        certify (bool, optional): boolean representing if the instruction is to
            certify (True) or decertify (False); defaults to True.

    Returns:
        HTTP response object returned by the MicroStrategy REST server.
    """
    url = (
        f'{connection.base_url}/api/objects/{id}/certify/?type={str(object_type)}'
        f'&certify={str(certify)}'
    )
    return connection.put(
        url=url,
        headers={'Content-Type': 'application/json', 'Accept': 'application/json'},
    )



def _jupyter_nbextension_paths():
    return [
        dict(
            section="notebook",
            src="production/mstr_jupyter/static",
            dest="mstr_jupyter",
            require="mstr_jupyter/main",
        )
    ]



# flake8: noqa
from .cube import list_all_cubes, load_cube
from .cube_cache import (
    CubeCache,
    delete_cube_cache,
    delete_cube_caches,
    list_cube_caches,
)
from .olap_cube import list_olap_cubes, OlapCube
from .super_cube import (
    list_super_cubes,
    SuperCube,
    SuperCubeAttribute,
    SuperCubeAttributeForm,
    SuperCubeFormExpression,
)



from typing import Dict, List

from mstrio.utils.dict_filter import filter_list_of_dicts


class MstrException(Exception):
    """Base class for exceptions returned by the MicroStrategy REST API.

    Attributes:
        code: Error code
        message: Error message
        ticket_id: MSTR Ticket ID
    """

    def __init__(self, err_data: Dict):
        self.code = err_data.get("code")
        self.message = err_data.get("message")
        self.ticket_id = err_data.get("ticketId")
        self.iserver_code = err_data.get("iServerCode")
        self.full_message = (
            f"{self.code}: {self.message} (Ticket ID: {self.ticket_id}, "
            f"iServerCode: {self.iserver_code})"
        )
        super().__init__(self.full_message)


class MstrTimeoutError(MstrException):
    pass


class VersionException(Exception):
    pass


class IServerException(Exception):
    pass


class PromptedContentError(Exception):
    pass


class Success(Exception):
    """This error holds details about the requested operation.

    Attributes:
        succeeded: list of succeeded operations dict elements
    """

    def __init__(self, data: List[dict]):
        assert isinstance(data, list)

        self.succeeded = data
        self.full_message = (
            f"Operation successful:\n{len(self.succeeded)} succeeded requests"
        )
        super().__init__(self.full_message)

    def __bool__(self):
        return True


class PartialSuccess(Exception):
    """This error holds details about the requested operation.

    Attributes:
        succeeded: list of succeeded operations dict elements
        failed: list of failed operations dict elements
    """

    def __init__(self, data: List[dict]):
        assert isinstance(data, list)

        self.succeeded = filter_list_of_dicts(data, status=[200, 204])
        self.failed = filter_list_of_dicts(data, status=">=400")
        self.full_message = (
            f"Operation partially successful:\n{len(self.failed)} failed "
            f"requests\n{len(self.succeeded)} succeeded requests"
        )
        super().__init__(self.full_message)

    def __bool__(self):
        return False



from typing import List, Optional

from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg='Error reading lock status of the schema.')
def read_lock_status(connection: "Connection", project_id: Optional[str] = None):
    """Read lock status of the schema."""
    project_id = project_id if project_id is not None else connection.project_id
    return connection.get(
        url=f"{connection.base_url}/api/model/schema/lock",
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg="Error placing the lock of type `{lock_type}` on the schema.")
def lock_schema(
    connection: "Connection",
    lock_type: str,
    project_id: Optional[str] = None,
    throw_error: bool = True,
):
    """Places a lock on the schema."""
    project_id = project_id if project_id is not None else connection.project_id
    return connection.post(
        url=f"{connection.base_url}/api/model/schema/lock",
        headers={'X-MSTR-ProjectID': project_id},
        json={'lockType': lock_type},
    )


@ErrorHandler(err_msg='Error unlocking the schema.')
def unlock_schema(
    connection: "Connection",
    lock_type: Optional[str] = None,
    project_id: Optional[str] = None,
    throw_error: bool = True,
):
    """Unlocks the schema."""
    project_id = project_id if project_id is not None else connection.project_id
    return connection.delete(
        url=f"{connection.base_url}/api/model/schema/lock",
        headers={'X-MSTR-ProjectID': project_id},
        params={'lockType': lock_type},
    )


@ErrorHandler(err_msg='Error reloading the schema.')
def reload_schema(
    connection: "Connection",
    project_id: Optional[str] = None,
    update_types: Optional[List[str]] = None,
    prefer_async: bool = False,
):
    """Reloads (updates) the schema."""
    project_id = project_id if project_id is not None else connection.project_id
    update_types = update_types if update_types else []
    prefer_async = 'respond-async' if prefer_async else None
    return connection.post(
        url=f"{connection.base_url}/api/model/schema/reload",
        headers={
            'X-MSTR-ProjectID': project_id,
            'Prefer': prefer_async,
        },
        json={'updateTypes': update_types},
    )


@ErrorHandler(err_msg='Error reading status of the task with ID: {task_id}.')
def read_task_status(
    connection: "Connection", task_id: str, project_id: Optional[str] = None
):
    """Read the status of the task."""
    project_id = project_id if project_id is not None else connection.project_id
    return connection.get(
        url=f"{connection.base_url}/api/model/tasks/{task_id}",
        headers={'X-MSTR-ProjectID': project_id},
    )



from datetime import datetime
from enum import auto
from typing import Optional, Union

from mstrio.utils.enum_helper import AutoName, AutoUpperName
from mstrio.utils.helper import Dictable, camel_to_snake, exception_handler


class SendContentAs(AutoName):
    DATA = auto()
    DATA_AND_HISTORY_LIST = auto()
    DATA_AND_LINK_AND_HISTORY_LIST = auto()
    LINK_AND_HISTORY_LIST = auto()


class Orientation(AutoUpperName):
    PORTRAIT = auto()
    LANDSCAPE = auto()


class CacheType(AutoName):
    RESERVED = auto()
    SHORTCUT = auto()
    SHORTCUTWITHBOOKMARK = 'shortcut_and_bookmark'


class LegacyCacheType(AutoUpperName):
    """
    Available values when using environment with version 11.3.0100 or lower
    """

    RESERVED = auto()
    SHORTCUT = auto()
    SHORTCUTWITHBOOKMARK = auto()
    BOOKMARK = auto()


class ShortcutCacheFormat(AutoUpperName):
    RESERVED = auto()
    JSON = auto()
    BINARY = auto()
    BOTH = auto()


class ClientType(AutoUpperName):
    RESERVED = auto()
    BLACKBERRY = auto()
    PHONE = auto()
    TABLET = auto()
    ANDROID = auto()


class LibraryCacheTypes(AutoName):
    ANDROID = auto()
    ANDROID_AND_IOS = auto()
    IOS = auto()
    WEB = auto()


class DeliveryDictable(Dictable):
    VALIDATION_DICT = {}

    @classmethod
    def from_dict(cls, source, **kwargs):
        """Initialize Delivery object from dictionary."""
        obj = cls.__new__(cls)
        super(DeliveryDictable, obj).__init__()
        source = camel_to_snake(source)
        for key, value in source.items():
            if key == 'zip':
                setattr(obj, key, ZipSettings.from_dict(value))
            else:
                setattr(obj, key, value)
        return obj

    def validate(self):
        """Validate whether all obligatory properties of the Delivery object
        are present and whether all the properties present are of
        correct types."""
        for key, value in self.__dict__.items():
            vtype = self.VALIDATION_DICT[key][0]
            obligatory = self.VALIDATION_DICT[key][1]
            if value and not isinstance(value, vtype):
                exception_handler(
                    "{} has incorrect type {}. Correct type is {}.".format(
                        key, type(value), vtype
                    ),
                    TypeError,
                )
            elif value is None and obligatory:
                exception_handler(
                    f"{key} is obligatory and cannot be empty.", ValueError
                )


class ZipSettings(DeliveryDictable):
    """Optional compression settings

    Attributes:
        filename: Filename of the compressed content
        password: Optional password for the compressed file
        password_protect: Whether to password protect file or not
    """

    VALIDATION_DICT = {
        "filename": [str, False],
        "password": [str, False],
        "password_protect": [bool, False],
    }

    def __init__(
        self,
        filename: Optional[str] = None,
        password: Optional[str] = None,
        password_protect: bool = False,
    ):
        self.filename = filename
        self.password = password if password_protect else None
        self.password_protect = password_protect
        self.validate()


class Delivery(DeliveryDictable):
    """Delivery settings object

    Attributes:
        mode: The subscription delivery mode (i.e. email, file, printer, etc.)
        expiration: Expiration date of the subscription, format should
            be yyyy-MM-dd
        contact_security: Whether to use contact security for each contact
            group member
        email: Email delivery properties object
        file: File delivery properties object
        printer: File delivery properties object
        ftp: FTP delivery properties object
        cache: Cache delivery properties
        mobile: Mobile delivery properties object
        history_list: HistoryList delivery properties
    """

    class DeliveryMode(AutoUpperName):
        EMAIL = auto()
        FILE = auto()
        PRINTER = auto()
        HISTORY_LIST = auto()
        CACHE = auto()
        MOBILE = auto()
        FTP = auto()
        SNAPSHOT = auto()
        PERSONAL_VIEW = auto()
        SHARED_LINK = auto()
        UNSUPPORTED = auto()

    class Email(DeliveryDictable):
        """Delivery properties for Email subscriptions

        Attributes:
            subject: The email subject associated with the subscription
            message: The email body of subscription
            filename: The filename that will be delivered when the subscription
                is executed
            space_delimiter: The space delimiter
            send_content_as: Send subscribed content as one of [data,
                data_and_history_list, data_and_link_and_history_list,
                link_and_history_list]
            overwrite_older_version: Whether the current subscription will
                overwrite earlier versions of the same report or document
                in the history list
            zip: Optional compression settings object
        """

        VALIDATION_DICT = {
            "subject": [str, True],
            "message": [str, False],
            "filename": [str, False],
            "space_delimiter": [str, False],
            "send_content_as": [str, False],
            "overwrite_older_version": [bool, False],
            "zip": [ZipSettings, False],
        }

        def __init__(
            self,
            subject: Optional[str] = None,
            message: Optional[str] = None,
            filename: Optional[str] = None,
            space_delimiter: Optional[str] = None,
            send_content_as: Optional[SendContentAs] = None,
            overwrite_older_version: bool = False,
            zip: Optional[ZipSettings] = None,
        ):
            self.subject = subject
            self.message = message
            self.filename = filename
            self.space_delimiter = space_delimiter
            self.send_content_as = send_content_as
            self.overwrite_older_version = overwrite_older_version
            self.zip = zip
            self.validate()

        def validate(self):
            """Validate whether all obligatory properties of the Delivery
            object are present, whether all the properties present are
            of correct types and whether the message and subject do
            not exceed the character limits."""
            if self.message and len(self.message) > 1000:
                exception_handler(
                    "Message too long. Max message length is 1000 characters."
                )
            if self.subject and len(self.subject) > 265:
                exception_handler(
                    "Subject too long. Max subject length is 265 characters."
                )
            super().validate()

    class File(DeliveryDictable):
        """Delivery properties for File subscriptions

        Attributes:
            filename: The filename that will be delivered when the subscription
                is executed
            space_delimiter: The space delimiter
            burst_sub_folder:The burst sub folder
            zip: Optional compression settings object
        """

        VALIDATION_DICT = {
            "filename": [str, False],
            "space_delimiter": [str, False],
            "burst_sub_folder": [str, False],
            "zip": [ZipSettings, False],
        }

        def __init__(
            self,
            filename: Optional[str] = None,
            space_delimiter: Optional[str] = None,
            burst_sub_folder: Optional[str] = None,
            zip: Optional[ZipSettings] = None,
        ):
            self.filename = filename
            self.space_delimiter = space_delimiter
            self.burst_sub_folder = burst_sub_folder
            self.zip = zip

    class Printer(DeliveryDictable):
        """Delivery properties for Printer subscriptions

        Attributes:
            copies: The number of copies that should be printed
            range_start: The number indicating the first report page that
                should be printed
            range_end: The number indicating the last report page that should
                be printed
            collated: Whether the printing should be collated or not
            orientation: Whether orientation is portrait or landscape
            use_print_range: Whether a print range should be used
        """

        VALIDATION_DICT = {
            "copies": [int, False],
            "range_start": [int, False],
            "range_end": [int, False],
            "collated": [bool, False],
            "orientation": [str, False],
            "use_print_range": [bool, False],
        }

        def __init__(
            self,
            copies: Optional[int] = None,
            range_start: Optional[int] = None,
            range_end: Optional[int] = None,
            collated: bool = False,
            orientation: Orientation = Orientation.PORTRAIT.name,
            use_print_range: bool = False,
        ):
            self.copies = copies
            self.range_start = range_start
            self.range_end = range_end
            self.collated = collated
            self.orientation = orientation
            self.use_print_range = use_print_range

    class Ftp(DeliveryDictable):
        """Delivery properties for FTP subscriptions

        Attributes:
            filename: The filename that will be delivered when the subscription
                is executed
            space_delimiter: The space delimiter
            zip: Optional compression settings object
        """

        VALIDATION_DICT = {
            "filename": [str, False],
            "space_delimiter": [str, False],
            "zip": [ZipSettings, False],
        }

        def __init__(
            self,
            space_delimiter: Optional[str] = None,
            filename: Optional[str] = None,
            zip: Optional[ZipSettings] = None,
        ):
            self.space_delimiter = space_delimiter
            self.filename = filename
            self.zip = zip

    class Cache(DeliveryDictable):
        """Delivery properties for Cache subscriptions

        Attributes:
            cache_type: The cache type to use
            shortcut_cache_format: The shortcut cache format to use
            library_cache_types: Set of library cache types,
                available types can be web, android, ios
            reuse_dataset_cache: Whether to reuse dataset cache
            is_all_library_users: Whether for all library users
        """

        VALIDATION_DICT = {
            "library_cache_types": [list, False],
            "reuse_dataset_cache": [bool, False],
            "is_all_library_users": [bool, False],
        }

        def __init__(
            self,
            cache_type: CacheType = CacheType.RESERVED,
            shortcut_cache_format: ShortcutCacheFormat = ShortcutCacheFormat.RESERVED,
            library_cache_types: Optional[list[LibraryCacheTypes]] = None,
            reuse_dataset_cache: bool = False,
            is_all_library_users: bool = False,
        ):
            self.cache_type = cache_type
            self.shortcut_cache_format = shortcut_cache_format
            self.library_cache_types = library_cache_types
            self.reuse_dataset_cache = reuse_dataset_cache
            self.is_all_library_users = is_all_library_users

    class Mobile(DeliveryDictable):
        """Delivery properties for Mobile subscriptions

        Attributes:
            client_type: The mobile client type
            device_id: The mobile target project
            do_not_create_update_caches: Whether the subscription will use
                a existing cache
            overwrite_older_version: Whether the current subscription will
                overwrite earlier versions of the same report or document in
                the history list
            re_run_hl: Whether the subscription will re-run against warehouse
        """

        VALIDATION_DICT = {
            "client_type": [str, False],
            "device_id": [str, False],
            "do_not_create_update_caches": [bool, False],
            "overwrite_older_version": [bool, False],
            "re_run_hl": [bool, False],
        }

        def __init__(
            self,
            client_type: ClientType = ClientType.RESERVED.name,
            device_id: Optional[str] = None,
            do_not_create_update_caches: bool = False,
            overwrite_older_version: bool = False,
            re_run_hl: bool = False,
        ):
            self.client_type = client_type
            self.device_id = device_id
            self.do_not_create_update_caches = do_not_create_update_caches
            self.overwrite_older_version = overwrite_older_version
            self.re_run_hl = re_run_hl

    class HistoryList(DeliveryDictable):
        """Delivery properties for History List subscriptions

        Attributes:
            device_id: The mobile target project
            do_not_create_update_caches: Whether the subscription will use
                a existing cache
            overwrite_older_version: Whether the current subscription will
                overwrite earlier versions of the same report or document in
                the history list
            re_run_hl: Whether the subscription will re-run against warehouse
        """

        VALIDATION_DICT = {
            "device_id": [str, False],
            "do_not_create_update_caches": [bool, False],
            "overwrite_older_version": [bool, False],
            "re_run_hl": [bool, False],
        }

        def __init__(
            self,
            device_id: Optional[str] = None,
            do_not_create_update_caches: bool = False,
            overwrite_older_version: bool = False,
            re_run_hl: bool = False,
        ):
            self.device_id = device_id
            self.do_not_create_update_caches = do_not_create_update_caches
            self.overwrite_older_version = overwrite_older_version
            self.re_run_hl = re_run_hl

    VALIDATION_DICT = {
        "mode": [str, True],
        "expiration": [str, False],
        "expiration_time_zone": [str, False],
        "contact_security": [bool, False],
        "email": [Email, False],
        "file": [File, False],
        "printer": [Printer, False],
        "ftp": [Ftp, False],
        "cache": [Cache, False],
        "mobile": [Mobile, False],
        "history_list": [HistoryList, False],
    }

    def __init__(
        self,
        mode=DeliveryMode.EMAIL.value,
        expiration: str = None,
        contact_security: Optional[bool] = None,
        subject: Optional[str] = None,
        message: Optional[str] = None,
        filename: Optional[str] = None,
        compress: bool = False,
        zip: Optional[ZipSettings] = None,
        password: Optional[str] = None,
        password_protect: bool = False,
        space_delimiter: Optional[str] = None,
        send_content_as: SendContentAs = SendContentAs.DATA,
        overwrite_older_version: bool = False,
        burst_sub_folder: Optional[str] = None,
        copies: Optional[int] = None,
        range_start: Optional[int] = None,
        range_end: Optional[int] = None,
        collated: bool = False,
        orientation: Orientation = Orientation.PORTRAIT.name,
        use_print_range: bool = False,
        client_type: ClientType = ClientType.RESERVED.name,
        device_id: Optional[str] = None,
        do_not_create_update_caches: bool = False,
        re_run_hl: bool = False,
        expiration_time_zone: Optional[str] = None,
        email: Optional[Email] = None,
        file: Optional[File] = None,
        cache_type: Union[CacheType, str] = CacheType.RESERVED,
        shortcut_cache_format: Union[
            ShortcutCacheFormat, str
        ] = ShortcutCacheFormat.RESERVED,
        library_cache_types: list[Union[LibraryCacheTypes, str]] = [
            LibraryCacheTypes.WEB
        ],
        reuse_dataset_cache: bool = False,
        is_all_library_users: bool = False,
        notification_enabled: bool = False,
        personal_notification_address_id: Optional[str] = None,
    ):
        self.mode = mode
        if expiration:
            if not self._expiration_check(expiration):
                exception_handler(
                    "Expiration date must be later or equal to {}".format(
                        datetime.now().strftime("%Y-%m-%d")
                    ),
                    ValueError,
                )
            else:
                self.expiration = expiration
        self.contact_security = contact_security
        self.expiration_time_zone = expiration_time_zone

        if zip is None and compress:
            temp_zip = ZipSettings(filename, password, password_protect)
        elif zip is not None and compress:
            temp_zip = zip
        else:
            temp_zip = None

        if self.DeliveryMode(mode) == self.DeliveryMode.EMAIL:
            self.email = self.Email(
                subject,
                message,
                filename,
                space_delimiter,
                send_content_as,
                overwrite_older_version,
                temp_zip,
            )
        elif self.DeliveryMode(mode) == self.DeliveryMode.FILE:
            self.file = self.File(filename, space_delimiter, burst_sub_folder, temp_zip)
        elif self.DeliveryMode(mode) == self.DeliveryMode.PRINTER:
            self.printer = self.Printer(
                copies, range_start, range_end, collated, orientation, use_print_range
            )
        elif self.DeliveryMode(mode) == self.DeliveryMode.FTP:
            self.ftp = self.Ftp(space_delimiter, filename, temp_zip)
        elif self.DeliveryMode(mode) == self.DeliveryMode.CACHE:
            self.cache = self.Cache(
                cache_type,
                shortcut_cache_format,
                library_cache_types,
                reuse_dataset_cache,
                is_all_library_users,
            )
        elif self.DeliveryMode(mode) == self.DeliveryMode.MOBILE:
            self.mobile = self.Mobile(
                client_type,
                device_id,
                do_not_create_update_caches,
                overwrite_older_version,
                re_run_hl,
            )
        elif self.DeliveryMode(mode) == self.DeliveryMode.HISTORY_LIST:
            self.history_list = self.HistoryList(
                device_id,
                do_not_create_update_caches,
                overwrite_older_version,
                re_run_hl,
            )
        if notification_enabled:
            self.notification_enabled = notification_enabled
            self.personal_notification = {
                "address_id": personal_notification_address_id
            }

    @classmethod
    def from_dict(cls, source, **kwargs):
        """Initialize Delivery object from dictionary."""
        obj = cls.__new__(cls)
        super(Delivery, obj).__init__()
        source = camel_to_snake(source)
        for key, value in source.items():
            if not obj.change_mode(
                key, value
            ):  # Change mode or set attr if its not mode param
                setattr(obj, key, value)
        return obj

    def change_mode(self, mode_name: str, mode_value) -> bool:
        """Change mode of the Delivery object.
        Return True on success, False if `mode_name` was invalid.
        """
        if mode_name == 'email':
            self.__setattr__(mode_name, self.Email.from_dict(mode_value))
        elif mode_name == 'file':
            self.__setattr__(mode_name, self.File.from_dict(mode_value))
        elif mode_name == 'mobile':
            self.__setattr__(mode_name, self.Mobile.from_dict(mode_value))
        elif mode_name == 'ftp':
            self.__setattr__(mode_name, self.Ftp.from_dict(mode_value))
        elif mode_name == 'cache':
            self.__setattr__(mode_name, self.Cache.from_dict(mode_value))
        elif mode_name == 'history_list':
            self.__setattr__(mode_name, self.HistoryList.from_dict(mode_value))
        else:
            return False
        return True

    def _expiration_check(self, date):
        now = datetime.now().date()
        date = datetime.strptime(date, '%Y-%m-%d').date()

        return date >= now



# flake8: noqa
from .base_subscription import Subscription
from .cache_update_subscription import CacheUpdateSubscription
from .content import Content
from .delivery import (
    CacheType,
    ClientType,
    Delivery,
    Orientation,
    SendContentAs,
    ShortcutCacheFormat,
    ZipSettings,
)
from .dynamic_recipient_list import DynamicRecipientList, list_dynamic_recipient_lists
from .email_subscription import EmailSubscription
from .file_subscription import FileSubscription
from .ftp_subscription import FTPSubscription
from .history_list_subscription import HistoryListSubscription
from .subscription_manager import list_subscriptions, SubscriptionManager



from enum import auto, Enum
import functools
from itertools import filterfalse
import logging
import sys
from textwrap import dedent
from typing import Optional, Union

from packaging.version import Version

from .exceptions import NotSupportedError
from .. import __version__, config

logger = logging.getLogger(__name__)
logging.basicConfig(  # NOSONAR
    level=logging.INFO, format='%(message)s', stream=sys.stderr
)
current_version = Version(__version__)


class WipLevels(Enum):
    SILENT = auto()
    PREVIEW = auto()
    INFO = auto()
    WARNING = auto()
    ERROR = auto()


def _wiplevel_error(level):
    raise ValueError(f'Unrecognised WipLevel {level}.')


message_templates = {
    WipLevels.PREVIEW: (
        "In a current version, {}is available as "
        "a Functionality Preview. It is subject to change until it is released "
        "as Generally Available."
    ),
    WipLevels.INFO: (
        "This {}functionality is a work-in-progress. It may change in future updates."
    ),
    WipLevels.WARNING: (
        "This {}functionality is a work-in-progress. Use it only if you understand"
        " the underlying code. It may change in future updates."
    ),
    WipLevels.ERROR: (
        "This {}functionality is a work-in-progress. It has been deemed unsafe to use"
        " and is currently blocked."
    ),
}
release_info_template = " It is currently scheduled for {} release."
docstring_prefix_template = dedent(
    """\
    ------------------ WORK IN PROGRESS ------------------

    {}

    ------------------------------------------------------

"""
)


def _get_message_template(level: WipLevels = WipLevels.WARNING):
    if level in message_templates.keys():
        return message_templates[level]
    else:
        return message_templates[WipLevels.INFO]


def _construct_message(
    name: Optional[str] = None,
    target_release: Optional[Union[Version, str]] = None,
    level: WipLevels = WipLevels.WARNING,
):
    template = _get_message_template(level)
    if name is None:
        message = template.format("")
    else:
        message = template.format(f"({name}) ")
    if target_release is not None:
        message += release_info_template.format(target_release)
    return message


def _emit(
    name: Optional[str] = None,
    target_release: Optional[Union[Version, str]] = None,
    level: WipLevels = WipLevels.WARNING,
    message: Optional[str] = None,
):
    if level == WipLevels.SILENT or not config.wip_warnings_enabled:
        return None

    if message is None:
        message = _construct_message(name, target_release, level)

    if level == WipLevels.INFO or level == WipLevels.PREVIEW:
        logger.info(message)
    elif level == WipLevels.WARNING:
        logger.warning(message)
    elif level == WipLevels.ERROR:
        raise NotSupportedError(message)
    else:
        _wiplevel_error(level)


def wip(
    target_release: Optional[Union[Version, str]] = None,
    level: WipLevels = WipLevels.WARNING,
    message: Optional[str] = None,
    prefix_doc: Union[bool, str] = True,
    mark_attr: bool = True,
):
    """Work-In-Progress wrapper

    Note:
        This is a decorator generator, which accepts arguments and returns the
        actual decorator, so even when not providing arguments and letting it
        choose the defaults it should be used like `@wip()`, not plain `@wip`.

    Args:
        target_release: The target release when the functionality is scheduled
            to be production-ready.
        level: The severity level of the warning emitted when the functionality
            is loaded or used.
            SILENT: no warning will be emitted
            PREVIEW: a warning will be dispatched
            INFO: a disclaimer will be printed to stderr
            WARNING: a warning will be dispatched
            ERROR: an error will be raised
        message: A custom message to replace the one standard to the given level
        prefix_doc: Whether to prefix the docstring with a WIP warning.
            Accepts bool values or a custom string to replace standard prefix.
        mark_attr: Marks the wrapped function via adding the `_wip` attribute,
            which is checked by `is_wip`.
    """
    if target_release is not None:
        if not isinstance(target_release, Version):
            target_release = Version(target_release)
        if target_release <= current_version:
            raise ValueError(
                "WIP wrapper called with target_release equal to"
                " or lower than current release."
            )
    if level not in WipLevels:
        _wiplevel_error(level)

    def wrap_func(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            _emit(wrapped.__name__, target_release, level, message)
            return f(*args, **kwargs)

        if mark_attr:
            wrapped._wip = True  # This makes it trivial to programmatically check.
        if prefix_doc:
            if not wrapped.__doc__:
                wrapped.__doc__ = ""
            if isinstance(prefix_doc, str):
                wrapped.__doc__ = prefix_doc + wrapped.__doc__
            else:
                if message is not None:
                    docstring_prefix = docstring_prefix_template.format(message)
                else:
                    msg = _construct_message(target_release=target_release, level=level)
                    docstring_prefix = docstring_prefix_template.format(msg)
                wrapped.__doc__ = docstring_prefix + wrapped.__doc__
        return wrapped

    return wrap_func


def module_wip(
    module_globals: dict,
    target_release: Optional[Union[Version, str]] = None,
    level: WipLevels = WipLevels.WARNING,
    message: Optional[str] = None,
    prefix_doc: Union[bool, str] = True,
    mark_attr: bool = True,
):
    """Emit the WIP warning/error/info when the module is loaded."""
    if mark_attr:
        module_globals["_wip"] = True
    if prefix_doc:
        if not module_globals["__doc__"]:
            module_globals["__doc__"] = ""
        if isinstance(prefix_doc, str):
            module_globals["__doc__"] = prefix_doc + module_globals["__doc__"]
        else:
            if message is not None:
                docstring_prefix = docstring_prefix_template.format(message)
            else:
                msg = _construct_message(target_release=target_release, level=level)
                docstring_prefix = docstring_prefix_template.format(msg)
            module_globals["__doc__"] = docstring_prefix + module_globals["__doc__"]
    _emit(f"module {module_globals['__name__']}", target_release, level, message)


def pause_wip_warnings():
    previous_setting = config.wip_warnings_enabled
    config.wip_warnings_enabled = False
    return previous_setting


def resume_wip_warnings(previous_setting: bool):
    config.wip_warnings_enabled = previous_setting


def is_wip(obj) -> bool:
    return getattr(obj, "_wip", False)


def is_module_wip(module_globals: dict) -> bool:
    return module_globals.get("_wip", False)


def filter_wip(iterable, return_wip: bool = False):
    """Return a generator containing only non-WIP elements
        (or only WIP elements, depending on `return_wip` parameter.).

    Args:
        iterable: Any iterable containing elements that may or may not be WIP.
        return_wip: Whether to return only WIP results or the opposite.
            Default `False` (return only non-WIP).
    """
    if return_wip:
        return filter(is_wip, iterable)
    return filterfalse(is_wip, iterable)


def filter_wip_dict(dict: dict, return_wip: bool = False) -> dict:
    """Return a dict containing only non-WIP elements
        (or only WIP elements, depending on `return_wip` parameter.).

    Args:
        dict: Any dict containing elements that may or may not be WIP.
        return_wip: Whether to return only WIP results or the opposite.
            Default `False` (return only non-WIP).
    """
    if return_wip:
        return {key: val for key, val in dict.items() if is_wip(val)}
    return {key: val for key, val in dict.items() if not is_wip(val)}



import logging
from typing import Optional

import pandas as pd

from mstrio.api import cubes
from mstrio.connection import Connection
from mstrio.object_management.search_operations import SearchPattern, full_search
from mstrio.utils.entity import ObjectSubTypes, ObjectTypes
from mstrio.utils.helper import exception_handler, get_valid_project_id
from mstrio.utils.version_helper import class_version_handler

from .cube import _Cube, load_cube

logger = logging.getLogger(__name__)


def list_olap_cubes(
    connection: Connection,
    name: Optional[str] = None,
    search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
    project_id: Optional[str] = None,
    project_name: Optional[str] = None,
    to_dictionary: bool = False,
    limit: Optional[int] = None,
    **filters,
) -> list["OlapCube"] | list[dict]:
    """Get list of OlapCube objects or dicts with them. Optionally filter cubes
    by specifying 'name'.

    Optionally use `to_dictionary` to choose output format.

    Wildcards available for 'name':
        ? - any character
        * - 0 or more of any characters
        e.g. name = ?onny will return Sonny and Tonny

    Specify either `project_id` or `project_name`.
    When `project_id` is provided (not `None`), `project_name` is omitted.

    Note:
        When `project_id` is `None` and `project_name` is `None`,
        then its value is overwritten by `project_id` from `connection` object.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        name (string, optional): value the search pattern is set to, which
            will be applied to the names of olap cubes being searched
        search_pattern (SearchPattern enum or int, optional): pattern to search
            for, such as Begin With or Contains. Possible values are available
            in ENUM mstrio.object_management.SearchPattern.
            Default value is BEGIN WITH (4).
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        to_dictionary (bool, optional): If True returns dict, by default (False)
            returns OlapCube objects
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned.
        **filters: Available filter parameters: ['id', 'name', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'owner',
            'ext_type', 'view_media', 'certified_info']

    Returns:
        list with OlapCubes or list of dictionaries
    """
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=False if project_name else True,
    )

    objects_ = full_search(
        connection,
        object_types=ObjectSubTypes.OLAP_CUBE,
        project=project_id,
        name=name,
        pattern=search_pattern,
        limit=limit,
        **filters,
    )
    if to_dictionary:
        return objects_
    return [OlapCube.from_dict(obj_, connection) for obj_ in objects_]


@class_version_handler('11.3.0000')
class OlapCube(_Cube):
    """Manage single table cube - according to EnumDSSXMLObjectSubTypes its
    subtype is  776 (DssXmlSubTypeReportCube). It inherits all properties from
    Cube.

    Attributes:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`.
        id: Identifier of a pre-existing cube.
        instance_id (str): Identifier of a cube instance if already initialized,
            None by default.
        name: name of the OlapCube
        size(integer): size of cube
        status(integer): status of cube
        path(string): full path of the cube on environment
        owner_id(string): ID of cube's owner
        attributes(list): all attributes of cube
        metrics(list): all metrics of cube
        attr_elements(list): all attributes elements of cube
        selected_attributes(list): IDs of filtered attributes
        selected_metrics(list): IDs of filtered metrics
        selected_attr_elements(list): IDs of filtered attribute elements
        dataframe(object): content of a cube extracted into a
            Pandas `DataFrame`
        table_definition
    """

    _OBJECT_SUBTYPE = ObjectSubTypes.OLAP_CUBE.value

    def __init__(
        self,
        connection: Connection,
        id: str,
        name: Optional[str] = None,
        instance_id: Optional[str] = None,
        parallel: bool = True,
        progress_bar: bool = True,
    ):
        """Initialize an Olap Cube instance.

        Note:
            Parameter `name` is not used when fetching. `id` is always used to
            uniquely identify cube.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            id (str): Identifier of a pre-existing cube.
            name (str): Name of a cube.
            instance_id (str): Identifier of an instance if cube instance has
                already been initialized, None by default.
            parallel (bool, optional): If True (default), utilize optimal number
                of threads to increase the download speed. If False, this
                feature will be disabled.
            progress_bar(bool, optional): If True (default), show the download
                progress bar.
        """
        # NOTE use the super.init from _Cube once cube_id is deprecated
        super().__init__(
            connection,
            id=id,
            name=name,
            instance_id=instance_id,
            parallel=parallel,
            progress_bar=progress_bar,
        )

    @classmethod
    def available_metrics(
        cls,
        connection: Connection,
        basic_info_only: bool = True,
        to_dataframe: bool = False,
    ) -> list[dict] | list[pd.DataFrame]:
        """Get all metrics available on I-Server.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`
            basic_info_only(boolean, optional): When True (default value) only
                values of `id`, `name` and `type` will be returned for each
                metric. When False, then all properties of each metric will be
                returned.
            to_dataframe: When False (default value) then metrics are returned
                as a list of dictionaries. When True then metrics are returned
                as Pandas 'DataFrame'.

        Returns:
            List of attributes or attributes as Pandas `DataFrame`.
        """
        return cls.__available_objects(
            connection, ObjectTypes.METRIC, basic_info_only, to_dataframe
        )

    @classmethod
    def available_attributes(
        cls,
        connection: Connection,
        basic_info_only: bool = True,
        to_dataframe: bool = False,
    ) -> list[dict] | list[pd.DataFrame]:
        """Get all attributes available on I-Server.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`
            basic_info_only(boolean, optional): When True (default value) only
                values of `id`, `name` and `type` will be returned for each
                attribute. When False, then all properties of each attribute
                will be returned.
            to_dataframe: When False (default value) then attributes are
                returned as a list of dictionaries. When True then attributes
                are returned as Pandas 'DataFrame'.

        Returns:
            List of attributes or attributes as Pandas `DataFrame`.
        """
        return cls.__available_objects(
            connection, ObjectTypes.ATTRIBUTE, basic_info_only, to_dataframe
        )

    @classmethod
    def available_attribute_forms(
        cls,
        connection: Connection,
        basic_info_only: bool = True,
        to_dataframe: bool = False,
    ) -> list[dict] | list[pd.DataFrame]:
        """Get all attribute forms available on I-Server.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`
            basic_info_only(boolean, optional): When True (default value) only
                values of `id`, `name` and `type` will be returned for each
                attribute form. When False, then all properties of each
                attribute form will be returned.
            to_dataframe: When False (default value) then attribute forms are
                returned as a list of dictionaries. When True then attribute
                forms are returned as Pandas 'DataFrame'.

        Returns:
            List of attribute forms or attribute forms as Pandas `DataFrame`.
        """
        return cls.__available_objects(
            connection, ObjectTypes.ATTRIBUTE_FORM, basic_info_only, to_dataframe
        )

    @classmethod
    def __available_objects(
        cls,
        connection: Connection,
        object_type: ObjectTypes | ObjectSubTypes,
        basic_info_only: bool = True,
        to_dataframe: bool = False,
    ) -> list[dict] | list[pd.DataFrame]:
        """Helper function to get available objects based on their type. It
        should be used to get only available attribute, metrics or attribute
        forms."""
        connection._validate_project_selected()
        avl_objects = full_search(
            connection=connection,
            object_types=object_type,
            project=connection.project_id,
        )
        for a in avl_objects:
            new_type = None
            if a['type'] == ObjectTypes.METRIC.value:
                new_type = 'metric'
            elif a['type'] == ObjectTypes.ATTRIBUTE.value:
                new_type = 'attribute'
            elif a['type'] == ObjectTypes.ATTRIBUTE_FORM.value:
                new_type = 'form'
            if new_type:
                a['type'] = new_type

        if basic_info_only:
            avl_objects = [
                {'id': a['id'], 'name': a['name'], 'type': a['type']}
                for a in avl_objects
            ]
        if to_dataframe:
            avl_objects = pd.DataFrame.from_dict(avl_objects)

        return avl_objects

    @classmethod
    def create(
        cls,
        connection: "Connection",
        name: str,
        folder_id: str,
        description: Optional[str] = None,
        overwrite: bool = False,
        attributes: Optional[list[dict]] = None,
        metrics: Optional[list[dict]] = None,
    ) -> "OlapCube | None":
        """Create an OLAP Cube by defining its name, description, destination
        folder, attributes and metrics.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            name(string): OLAP Cube name.
            folder_id(string): Object ID of the folder where the cube should be
                saved.
            description(string, optional): OLAP Cube description
            overwrite(boolean, optional): Default value (False) not allow to
                overwrite the object with the same name.
            attributes(list of dicts, optional): list with dicts of attributes
                dicts to be in the working set of OLAP Cube. Each attribute dict
                should have keys: `id`, `name` and `type`. Attributes can be
                found with method `OlapCube.available_attributes`.
            metrics(list of dicts, optional): list with dicts of metrics to be
                in the working set of OLAP Cube. Each metric dict should have
                keys: `id`, `name` and `type`. Metrics can be found with
                method `OlapCube.available_metrics`.

        Returns:
            Newly created OLAP Cube or None in case of wrongly provided
            attributes or metrics.

        Raises:
            `requests.exceptions.HTTPError` when response returned from request
            to I-Server to create new OLAP Cube was not ok.
        """
        attributes = attributes or []
        metrics = metrics or []

        if not OlapCube.__check_objects(attributes, 'attribute'):
            return None
        if not OlapCube.__check_objects(metrics, 'metric'):
            return None

        definition = {
            'availableObjects': {'attributes': attributes, 'metrics': metrics}
        }
        cube_id = cubes.create(
            connection, name, folder_id, overwrite, description, definition
        ).json()['id']
        return load_cube(connection, cube_id)

    @staticmethod
    def __check_objects(objects_: list[dict], obj_name: str) -> bool:
        """Check objects (attribute or metrics) before creation or update of an
        OLAP Cube."""
        for obj in objects_:
            if not OlapCube.__check_object(obj, obj_name):
                return False
        return True

    @staticmethod
    def __check_object(object_: dict, obj_name: str) -> bool:
        """Check a single object (attribute or metric) represented as dictionary
        before creation or update of OLAP Cube. Each dictionary must have keys:
        `id` and `name`. When it has key `type` then it must be the same as
        value of parameter `obj_name`. When this key is missing its value will
        be set to value of parameter `obj_name`."""

        if 'id' not in object_:
            msg = f"{obj_name.capitalize()} '{object_['name']}' is missing key 'id'."
            exception_handler(msg, Warning, True)
            return False

        if 'type' not in object_:
            object_['type'] = obj_name
        elif object_['type'] != obj_name:
            msg = (
                f"Each element in dictionary with {obj_name}s must be of a type "
                f"'{obj_name}'."
            )
            exception_handler(msg, Warning, True)
            return False
        return True

    def update(self, attributes: list[dict] = [], metrics: list[dict] = []) -> bool:
        """Update an OLAP Cube. When Cube is unpublished, then it is possible to
         add or remove attributes and metrics to/from its definition and
         rearrange existing one. When cube is published it is possible only to
         rearrange attributes and metrics existing in its definition. After this
         operation cube will have only attributes and metrics provided in
         parameters.

         Args:
            attributes(list of dicts, optional): list with dicts of attributes
                dicts to be in the working set of OLAP Cube. Each attribute dict
                should have keys: `id`, `name` and `type`. Attributes can be
                found with method `OlapCube.available_attributes`.
            metrics(list of dicts, optional): list with dicts of metrics to be
                in the working set of OLAP Cube. Each metric dict should have
                keys: `id`, `name` and `type`. Metrics can be found with
                method `OlapCube.available_metrics`.

        Returns:
            True when update was successful. False otherwise.

        Raises:
            `requests.exceptions.HTTPError` when response returned from request
            to I-Server to update new OLAP Cube was not ok.
        """
        if not OlapCube.__check_attributes_update(
            attributes, self.attributes, self.status
        ):
            return False
        if not OlapCube.__check_metrics_update(metrics, self.metrics, self.status):
            return False

        definition = {
            'availableObjects': {'attributes': attributes, 'metrics': metrics},
        }

        res = cubes.update(self._connection, self._id, definition)
        # refresh definition of cube
        self._get_definition()
        return res.ok

    @staticmethod
    def __check_attributes_update(
        attributes: list[dict], existing_attributes: list[dict], status: int
    ) -> bool:
        """Check dictionaries with attributes before update of an OLAP Cube."""
        return OlapCube.__check_objects_update(
            attributes, existing_attributes, 'attribute', status
        )

    @staticmethod
    def __check_metrics_update(
        metrics: list[dict], existing_metrics: list[dict], status: int
    ) -> bool:
        """Check dictionaries with metrics before update of an OLAP Cube."""
        return OlapCube.__check_objects_update(
            metrics, existing_metrics, 'metric', status
        )

    @staticmethod
    def __check_objects_update(
        objects_: list[dict],
        existing_objects: list[dict],
        object_name: str,
        status: int,
    ) -> bool:
        """Check objects (attributes or metrics) represented as dictionaries
        before update of an OLAP Cube. When status of cube is 0, then it is not
        published and it is possible to freely add or delete objects. Otherwise
        it is possible to only rearrange existing objects. For each object it is
        done also the same check as before creation of an OLAP Cube.
        """
        existing_ids = [o['id'] for o in existing_objects]
        reorganised_objects_count = 0  # to check if all existing objects were provided
        msg = (
            f"It is not possible to add new {object_name}s when editing published cube."
        )
        for object_ in objects_:
            # check if structure of dictionary with an object is correct
            if not OlapCube.__check_object(object_, object_name):
                return False
            # check if status of cube is correct in case of new objects
            if object_['id'] not in existing_ids:
                if status != 0:
                    exception_handler(msg, Warning)
                    return False
            else:
                reorganised_objects_count += 1

        # check if status of cube is correct in case of removing objects
        if reorganised_objects_count != len(existing_ids) and status != 0:
            msg = (
                f"It is not possible delete existing {object_name}s when editing "
                f"published cube."
            )
            exception_handler(msg, Warning)
            return False

        return True

    def publish(self) -> None:
        """Publish an OLAP Cube. Request to publish an OLAP Cube is an
        asynchronous operation, so the result of it can be seen after calling
        method `refresh_status()` inherited from Cube class.

        Returns:
            A dictionary with two keys identifying task IDs.
        """
        response = cubes.publish(self._connection, self._id)
        logger.info(f"Request for publishing cube '{self.name}' was sent.")
        return response.json()

    def export_sql_view(self):
        """Export SQL View of an OLAP Cube.

        Returns:
            SQL View of an OLAP Cube.
        """
        res = cubes.get_sql_view(self._connection, self._id)
        sql_statement = res.json()['sqlStatement']
        return sql_statement



"""This is the demo script to show how to export data from MicroStrategy with
SuperCube. It is possible to create and publish single or multi-table SuperCube.

You can update super cube with different policies when adding table:
 - add      -> insert entirely new data
 - update   -> update existing data
 - upsert   -> simultaneously updates existing data and inserts new data
 - replace  -> truncates and replaces the data

You can set in how big increments data is transmitted to the server with the
`chunksize` parameter.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to ease
its usage.
"""

import pandas as pd

from mstrio.project_objects import (
    SuperCubeAttribute,
    SuperCubeAttributeForm,
    SuperCubeFormExpression
)
from mstrio.project_objects.datasets import SuperCube
from mstrio.connection import get_connection

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert project name here

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# prepare Pandas DataFrames to add it into tables of super cube
stores = {"store_id": [1, 2, 3], "location": ["New York", "Seattle", "Los Angeles"]}
stores_df = pd.DataFrame(stores, columns=["store_id", "location"])

sales = {
    "store_id": [1, 2, 3],
    "category": ["TV", "Books", "Accessories"],
    "sales": [400, 200, 100],
    "sales_fmt": ["$400", "$200", "$100"]
}
sales_df = pd.DataFrame(sales, columns=["store_id", "category", "sales", "sales_fmt"])

# Define a variable which can be later used in a script
SUPER_CUBE_NAME = $super_cube_name  # Insert name of created suber cube here

# Add tables to the super cube and create it. By default 'create()' will
# additionally upload data to the I-Server and publish it. You can manipulate it
# by setting parameters `auto_upload` and `auto_publish`
ds = SuperCube(connection=conn, name=SUPER_CUBE_NAME)
ds.add_table(name="Stores", data_frame=stores_df, update_policy="add")
ds.add_table(name="Sales", data_frame=sales_df, update_policy="add")
ds.create()

# Add tables to super cube and map columns to attribute forms
ds = SuperCube(connection=conn, name=SUPER_CUBE_NAME)
ds.add_table(name="Stores", data_frame=stores_df, update_policy="add")
ds.add_table(name="Sales", data_frame=sales_df, update_policy="add")

attribute_form_mapping = SuperCubeAttribute(
    name='store',
    forms=[
        SuperCubeAttributeForm(
            category='ID',
            expressions=[
                SuperCubeFormExpression(
                    table='Stores',
                    column='store_id'
                ),
                SuperCubeFormExpression(
                    table='Sales',
                    column='store_id'
                )
            ]
        ),
        SuperCubeAttributeForm(
            category='DESC',
            expressions=[
                SuperCubeFormExpression(
                    table='Stores',
                    column='name'
                )
            ]
        )
    ]
)

ds.create(attribute_forms=[attribute_form_mapping])

# When using `SuperCube.add_table()`, Pandas data types are mapped to
# MicroStrategy data types. By default, numeric data is modeled as MSTR metrics
# and non-numeric as attributes. You can set manually which columns treat as
# attributes and which as metrics.
ds.add_table(name="Stores", data_frame=stores_df, update_policy="add", to_attribute=["store_id"])

ds.add_table(
    name="Sales",
    data_frame=sales_df,
    update_policy="add",
    to_attribute=["store_id"],
    to_metric=["sales_fmt"]
)

# Define a variable which can be later used in a script
SUPER_CUBE_ID = $super_cube_id  # insert ID of edited super cube here

# It is possible to update previously created super cubes what looks really
# similar to creation. You can use different update policies which are explained
# in the description of this script at the top. By default, `update()`
# is publishing data automatically, if you don't want to publish data,
# you have to set argument 'auto_publish` to False. It is also possible to set
# chunksize for the update.
ds = SuperCube(connection=conn, id=SUPER_CUBE_ID)
ds.add_table(name="Stores", data_frame=stores_df, update_policy="update")
ds.add_table(name="Sales", data_frame=sales_df, update_policy="upsert")
ds.update()

# Finally, it is possible to certify an existing super cube
ds.certify()

# Limitations
# Updating SuperCubes that were not created using the MicroStrategy REST API is
# not possible. This applies for example to Cubes created via MicroStrategy Web
# client.



from dataclasses import dataclass
from enum import auto
import logging
from typing import Optional, TYPE_CHECKING

from requests import Response

from mstrio import config
from mstrio.api import monitors
from mstrio.connection import Connection
from mstrio.server import Cluster
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import camel_to_snake, Dictable, fetch_objects

if TYPE_CHECKING:
    from mstrio.project_objects import ContentCache

logger = logging.getLogger(__name__)

NO_CACHE_LOG = 'There are no caches meeting the criteria.'


class Cache:
    """
    Base class for managing cache.
    """

    def __init__(
        self, connection: Connection, cache_id: str, cache_dict: Optional[dict] = None
    ):
        """Initialize the Cache object.

        Args:
            connection (Connection): MicroStrategy connection object returned by
                `connection.Connection()`.
            cache_id (string): cache id
            cache_dict (dict, optional): dictionary with properties of cache
                object.
        """
        if not cache_dict:
            cache_dict = {}
        self._id = cache_id
        self._connection = connection
        self._project_id = connection.project_id

        self._init_variables(**cache_dict)

    def _init_variables(self, **kwargs) -> None:
        """Initialize variables given cache_dict."""
        kwargs = camel_to_snake(kwargs)
        self._project_name = kwargs.get('project_name')
        self._source = (
            CacheSource.from_dict(source) if (source := kwargs.get('source')) else None
        )
        self._last_update_time = kwargs.get('last_update_time')
        self._last_hit_time = kwargs.get('last_hit_time')
        self._hit_count = kwargs.get('hit_count')
        self._size = kwargs.get('size')
        self._creator_name = kwargs.get('creator_name') or kwargs.get('creator')
        self._creation_time = kwargs.get('creation_time')

    def list_properties(self):
        """List properties for cache."""
        return {
            'id': self.id,
            'size': self.size,
            'source': self.source,
            'last_update_time': self.last_update_time,
            'last_hit_time': self.last_hit_time,
            'hit_count': self.hit_count,
            'creator_name': self.creator_name,
            'creation_time': self.creation_time,
        }

    @property
    def id(self):
        return self._id

    @property
    def project_id(self):
        return self._project_id

    @property
    def source(self):
        return self._source

    @property
    def size(self):
        return self._size

    @property
    def last_update_time(self):
        return self._last_update_time

    @property
    def last_hit_time(self):
        return self._last_hit_time

    @property
    def hit_count(self):
        return self._hit_count

    @property
    def creator_name(self):
        return self._creator_name

    @property
    def creation_time(self):
        return self._creation_time


@dataclass
class CacheSource(Dictable):
    class Type(AutoName):
        REPORT = auto()
        DOCUMENT = auto()
        DOSSIER = auto()
        DOCDOSSIER = auto()
        CUBE = auto()

    _FROM_DICT_MAP = {
        'type': Type,
    }

    id: str
    name: str
    type: Type

    def __repr__(self):
        return (
            f"CacheSource(id='{self.id}, name='{self.name}, type='{self.type.value}')"
        )


class ContentCacheMixin:
    """ContentCacheMixin class adds ContentCache management for supporting cache

    Objects currently supported are ContentCache for documents, dossiers and
    reports.
    """

    @staticmethod
    def fetch_nodes(connection: "Connection", project_id: str) -> list[str]:
        """Fetches the nodes for the specified connection and project.

        Args:
            connection (Connection): MicroStrategy connection object returned
                by 'connection.Connection()'
            project_id (string): id of the project to fetch the nodes from

        Returns:
            A list of node names for the specified project."""
        cluster_ = Cluster(connection)
        nodes = cluster_.list_nodes(project=project_id, to_dictionary=True)
        nodes = [node.get('name') for node in nodes if node.get('name')]
        return nodes

    @staticmethod
    def __alter_status(
        connection: "Connection",
        op: str,
        cache_ids: list[str],
        value: Optional[bool] = None,
        status: Optional[str] = None,
        nodes: Optional[list[str]] = None,
    ) -> Response:
        """Engine for altering ContentCache status

        Args:
            connection (object): MicroStrategy connection object returned
                by `connection.Connection()`
            op (str): Replace or Remove operation to be performed
            cache_ids (list): list of cache ids on which the operation should be
                performed
            value (bool, optional): Value used by operation [new value used to
                replace existing value]
            status (str, optional): Status on which the value should be changed
            nodes (list, optional): list of node names for the specified project

        Returns:
            Response object
        """
        from mstrio.project_objects import ContentCache

        if not nodes:
            nodes = ContentCacheMixin.fetch_nodes(connection, connection.project_id)
        body = {'operationList': []}
        logger_message = {
            'replace/loaded/True': 'load',
            'replace/loaded/False': 'unload',
            'remove/None/None': 'delete',
        }.get(f'{op}/{status}/{value}')

        for cache_id in cache_ids:
            content_cache = ContentCache(connection, cache_id)
            if not content_cache.combined_id and config.verbose:
                logger.info(f'Could not {logger_message} cache {cache_id}')
                continue
            body['operationList'].append(
                {
                    'op': op,
                    'path': (
                        f'/contentCaches/{content_cache.combined_id}/status/{status}'
                    )
                    if status
                    else f'/contentCaches/{content_cache.combined_id}',
                    'value': value,
                }
            )
        if body['operationList']:
            return monitors.update_contents_caches(connection, nodes, body)
        elif config.verbose:
            logger.info(NO_CACHE_LOG)

    @staticmethod
    def load_caches(connection: "Connection", cache_ids: list[str]) -> Response:
        """Bulk load caches.

        Args:
            connection (Connection): MicroStrategy connection object returned
            by 'connection.Connection()'
            cache_ids (list[str]): list of cache ids to be loaded

        Returns:
            Response object."""
        res = ContentCacheMixin.__alter_status(
            connection=connection,
            op='replace',
            cache_ids=cache_ids,
            value=True,
            status='loaded',
        )
        if config.verbose and res:
            logger.info('Successfully loaded content caches')
        return res

    @staticmethod
    def unload_caches(connection: "Connection", cache_ids: list[str]) -> Response:
        """Bulk unload caches.

        Args:
            connection (Connection): MicroStrategy connection object returned
            by 'connection.Connection()'
            cache_ids (list[str]): list of cache ids to be unloaded

        Returns:
            Response object."""
        res = ContentCacheMixin.__alter_status(
            connection=connection,
            op='replace',
            cache_ids=cache_ids,
            value=False,
            status='loaded',
        )
        if config.verbose and res:
            logger.info('Successfully unloaded content caches')
        return res

    @staticmethod
    def delete_caches(
        connection: "Connection", cache_ids: list[str], force: Optional[bool] = None
    ) -> Response:
        """Bulk delete caches.

        Args:
            connection (Connection): MicroStrategy connection object returned
            by 'connection.Connection()'
            cache_ids (list[str]): list of cache ids to be deleted
            force (bool, optional): If True, then no additional prompt will be
                shown before deleting objects.

        Returns:
            Response object."""
        user_input = 'N'
        if not force:
            user_input = (
                input(
                    'Are you sure you want to delete all content caches with '
                    'provided IDs? [Y/N]: '
                )
                or 'N'
            )
        if force or user_input == 'Y':
            res = ContentCacheMixin.__alter_status(
                connection=connection, op='remove', cache_ids=cache_ids
            )
            if config.verbose and res:
                logger.info('Successfully deleted content caches')
            return res

    @classmethod
    def list_caches(
        cls,
        connection: "Connection",
        to_dictionary: bool = False,
        status: str = 'ready',
        project_id: Optional[str] = None,
        nodes: Optional[list[str] | str] = None,
        content_type: Optional[CacheSource.Type | str] = None,
        limit: Optional[int] = None,
        db_connection_id: Optional[str] = None,
        db_login_id: Optional[str] = None,
        id: Optional[str] = None,
        owner: Optional[str] = None,
        size: Optional[str] = None,
        wh_tables: Optional[str] = None,
        security_filter_id: Optional[str] = None,
    ) -> list["ContentCache"] | list[dict]:
        """List content caches. You can filter them by id, database
        connection (`db_connection_id`) and project (`project_id`).

        You can specify from which `nodes` caches will be retrieved. If `nodes`
        are `None` then all nodes are retrieved from the cluster.

        Args:
            connection (Connection): MicroStrategy connection object returned by
                `connection.Connection()`.
            to_dictionary (bool, optional): If True returns dict, by default
                (False) returns ContentCache objects
            status (string, optional): When provided, only caches with given
                status will be returned (if any). Default value `ready`
            project_id (string, optional): When provided only caches
                for project with given ID will be returned (if any).
            nodes (list[string] | string, optional): names of nodes on which
                caches will be searched. By default, it equals `None` and in
                that case all nodes names are loaded from the cluster.
            content_type (string | CacheSource.Type, optional): When provided,
                only caches of given type will be returned (if any).
            limit (integer, optional): Cut-off value for the number of objects
                returned. Default value is `1000` which is a maximum value.
            db_connection_id (string, optional): When provided, only caches for
                the database connection with given ID will be returned (if any).
            db_login_id (string, optional): When provided, only caches for
                the database login with given ID will be returned (if any).
            id (string, optional): When provided, only cache with
                given ID will be returned (if any).
            owner (string, optional): Owner of the content cache. Exact match on
                the owner's full name.
            size (string, optional): Size condition for the content cache
                (in KB). When provided, only caches which satisfy the condition
                will be returned (if any).
            wh_tables (string, optional):  When provided, only caches using
                given warehouse tables will be returned (if any).
            security_filter_id (string, optional): When provided, only caches
                using given security filter will be returned (if any).

        Returns:
            List of ContentCache objects when parameter `to_dictionary` is set
            to False (default value) or list of dictionaries otherwise.
        """
        if not project_id:
            project_id = connection.project_id
        if not nodes:
            nodes = ContentCacheMixin.fetch_nodes(connection, project_id)
        if not content_type:
            content_type = cls._CACHE_TYPE
        unloaded = False
        if status == 'unloaded':
            unloaded = True
            status = 'ready'

        caches = fetch_objects(
            connection=connection,
            api=monitors.get_contents_caches,
            project_id=project_id,
            node=nodes,
            limit=None,
            status=status,
            content_type=content_type,
            size=size,
            owner=owner,
            filters={},
        )
        caches = [list(cache.items())[0] for cache in caches['content_caches']]
        if limit:
            caches = caches[:limit]

        caches = [{**cache[1], 'combined_id': cache[0]} for cache in caches]

        # apply filtering
        def filtered_caches(str_arg, arg):
            return [cache for cache in caches if cache.get(str_arg, '') == arg]

        if id:
            caches = filtered_caches('id', id)
        if db_connection_id:
            caches = filtered_caches('databaseConnectionId', db_connection_id)
        if db_login_id:
            caches = filtered_caches('databaseLoginId', db_login_id)
        if wh_tables:
            caches = filtered_caches('warehouseTablesUsed', wh_tables)
        if security_filter_id:
            caches = filtered_caches('securityFilterId', security_filter_id)
        if unloaded:
            caches = [
                cache for cache in caches if not cache.get('status').get('loaded')
            ]

        if to_dictionary:
            return caches
        else:
            from mstrio.project_objects import ContentCache

            return ContentCache.from_dict(connection, caches)

    @classmethod
    def load_all_caches(
        cls,
        connection: "Connection",
        **filters,
    ):
        """
        Load all content caches filtered by the class type.
        Optionally filter by additional conditions.

        Args:
            cls (object): Class type for objects to be filtered by
            connection (Connection): MicroStrategy connection object returned by
                `connection.Connection()`
            **filters: Available filter parameters: ['db_connection_id',
                'db_login_id', 'owner', 'status', 'size', 'wh_tables',
                'security_filter_id']
        """
        cache_ids = [cache.id for cache in cls.list_caches(connection, **filters)]
        if cache_ids:
            cls.load_caches(
                connection=connection,
                cache_ids=cache_ids,
            )
        elif config.verbose:
            logger.info(NO_CACHE_LOG)

    @classmethod
    def unload_all_caches(
        cls,
        connection: "Connection",
        **filters,
    ):
        """
        Unload all content caches filtered by the class type.
        Optionally filter by additional conditions.

        Args:
            cls (object): Class type for objects to be filtered by
            connection (Connection): MicroStrategy connection object returned by
                `connection.Connection()`
            **filters: Available filter parameters: ['db_connection_id',
                'db_login_id', 'owner', 'status', 'size', 'wh_tables',
                'security_filter_id']
        """
        cache_ids = [cache.id for cache in cls.list_caches(connection, **filters)]
        if cache_ids:
            cls.unload_caches(
                connection=connection,
                cache_ids=cache_ids,
            )
        elif config.verbose:
            logger.info(NO_CACHE_LOG)

    @classmethod
    def delete_all_caches(
        cls,
        connection: "Connection",
        force: Optional[bool] = None,
        **filters,
    ):
        """
        Delete all content caches filtered by the class type.
        Optionally filter by additional conditions.

        Args:
            cls (object): Class type for objects to be filtered by
            connection (Connection): MicroStrategy connection object returned by
                `connection.Connection()`
            force (bool, optional): If True, then no additional prompt will be
                shown before deleting objects.
            **filters: Available filter parameters: ['db_connection_id',
                'db_login_id', 'owner', 'status', 'size', 'wh_tables',
                'security_filter_id']
        """
        cache_ids = [
            cache.id
            for cache in cls.list_caches(connection, **filters)
            if cache.status.ready
        ]
        if cache_ids:
            cls.delete_caches(
                connection=connection,
                cache_ids=cache_ids,
                force=force,
            )
        elif config.verbose:
            logger.info(NO_CACHE_LOG)



from enum import auto
import logging
from typing import List, Optional, TYPE_CHECKING, Union

from mstrio import config
from mstrio.api import datasources, objects
from mstrio.datasources.datasource_login import DatasourceLogin
from mstrio.users_and_groups.user import User
from mstrio.utils import helper
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, ObjectTypes
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.helper import (
    get_args_from_func,
    get_default_args_from_func,
    get_objects_id,
)
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.connection import Connection

logger = logging.getLogger(__name__)


@method_version_handler('11.3.0000')
def list_datasource_connections(
    connection: "Connection",
    to_dictionary: bool = False,
    limit: Optional[int] = None,
    **filters,
) -> Union[List["DatasourceConnection"], List[dict]]:
    """Get list of DatasourceConnection objects or dicts. Optionally filter the
    connections by specifying filters.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        to_dictionary: If True returns dict, by default (False) returns
            User objects.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['name', 'id', 'description',
            'acg', 'execution_mode', 'max_cancel_attempt_time',
            'max_query_exe_time', 'max_connection_attempt_time',
            'connection_lifetime', 'connection_idle_timeout',
            'char_encoding_windows', 'char_encoding_unix',
            'table_prefix', 'connection_string', 'parameterized_queries',
            'extended_fetch', 'driver_type', 'oauth_parameter', 'iam',
            'resource', 'scope', 'enable_sso', 'datasource_login',
            'database_type', 'database_version']

    Examples:
        >>> list_datasource_connections(connection, name='db_conn_name')
    """
    return DatasourceConnection._list_datasource_connections(
        connection=connection,
        to_dictionary=to_dictionary,
        limit=limit,
        **filters,
    )


class CharEncoding(AutoName):
    UTF8 = auto()
    NON_UTF8 = "multibyte"


class DriverType(AutoName):
    RESERVED = auto()
    ODBC = auto()
    NATIVE = auto()


class ExecutionMode(AutoName):
    RESERVED = auto()
    ASYNC_CONNECTION = auto()
    ASYNC_STATEMENT = auto()
    SYNCHRONOUS = auto()


@class_version_handler('11.3.0000')
class DatasourceConnection(Entity, CopyMixin, DeleteMixin):
    """Datasource connection configuration object that represents a connection
    to the datasource.

    Attributes:
        connection: A MicroStrategy connection object.
        id: Unique datasource connection ID.
        name: Unique datasource connection name.
        description: Datasource connection description.
        execution_mode: ExecutionMode Enum specifying how SQL statements will be
            executed on a physical database connection made.
        max_cancel_attempt_time: Number of seconds before being timed out when
            attempting to make a connection.
        max_query_exe_time: Number of seconds during which a query must be
            executed before being timed out.
        max_connection_attempt_time: Number of seconds connection attempts to be
            made before timing out.
        connection_lifetime: Number of seconds of the connection lifetime.
        connection_idle_timeout: Number of seconds before being timed out when
            the connection is idle.
        char_encoding_windows: CharEncoding Enum specifying the encoding
            across connection for Windows drivers.
        char_encoding_unix: CharEncoding Enum specifying the encoding across
            connection for Unix drivers.
        table_prefix: String prefixed to the names of all temporary tables
            created by the Query Engine for this connection during report
            execution with this string.
        connection_string: Database connection string.
        parameterized_queries: Specifies whether parameterized queries are
            enabled.
        extended_fetch: Specifies whether or not to use the extended fetch ODBC
            call to retrieve data from the database connection.
        datasource_login: `DatasourceLogin` object or ID
        database_type: Database type
        database_version: Database version
        driver_type: DriverType Enum specifying Driver used for database
            connection.
        oauth_parameter: Used for authentication with oAuth.
        type: Object type
        subtype: Object subtype
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        owner: User object that is the owner
        acg: Object access rights. It a bit vector where bits are defined at
            [EnumDSSXMLAccessRightFlags].
        acl: Object access control list
        iam: List of projects to which the fence is applied.
        resource: The url of configured Web API for OAuth authentication usage.
        scope: List of delegated permissions that the app is requesting.
        enable_sso: Specifies whether to use Single Sign-On.
    """

    _OBJECT_TYPE = ObjectTypes.DBCONNECTION
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': User.from_dict,
        'execution_mode': ExecutionMode,
        'datasource_login': DatasourceLogin.from_dict,
        'driver_type': DriverType,
        'char_encoding_unix': CharEncoding,
        'char_encoding_windows': CharEncoding,
    }
    _API_GETTERS = {
        (
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acl',
        ): objects.get_object_info,
        (
            'id',
            'name',
            'description',
            'execution_mode',
            'max_cancel_attempt_time',
            'max_query_exe_time',
            'max_connection_attempt_time',
            'connection_lifetime',
            'connection_idle_timeout',
            'char_encoding_windows',
            'char_encoding_unix',
            'table_prefix',
            'connection_string',
            'parameterized_queries',
            'extended_fetch',
            'driver_type',
            'datasource_login',
            'database_type',
            'database_version',
            'oauth_parameter',
            'acg',
            'iam',
            'resource',
            'scope',
            'enable_sso',
        ): datasources.get_datasource_connection,
    }
    _API_PATCH: dict = {
        ('abbreviation'): (objects.update_object, 'partial_put'),
        (
            "name",
            "description",
            "execution_mode",
            "max_cancel_attempt_time",
            "max_query_exe_time",
            "max_connection_attempt_time",
            "connection_lifetime",
            "connection_idle_timeout",
            "char_encoding_windows",
            "char_encoding_unix",
            "table_prefix",
            "connection_string",
            "parameterized_queries",
            "extended_fetch",
            "driver_type",
            "database_type",
            "database_version",
            "datasource_login",
            'iam',
            'resource',
            'scope',
            'enable_sso',
        ): (
            datasources.update_datasource_connection,
            'patch',
        ),
    }
    _PATCH_PATH_TYPES = {
        "name": str,
        "description": str,
        "execution_mode": str,
        "max_cancel_attempt_time": int,
        "max_query_exe_time": int,
        "max_connection_attempt_time": int,
        "connection_lifetime": int,
        "connection_idle_timeout": int,
        "char_encoding_windows": str,
        "char_encoding_unix": str,
        "table_prefix": str,
        "connection_string": str,
        "parameterized_queries": bool,
        "extended_fetch": bool,
        "driver_type": str,
        "database_type": str,
        "database_version": str,
        "datasource_login": dict,
        "iam": dict,
        "resource": str,
        "scope": str,
        "enable_sso": bool,
    }

    def __init__(
        self,
        connection: "Connection",
        name: Optional[str] = None,
        id: Optional[str] = None,
    ) -> None:
        """Initialize DatasourceConnection object and synchronize with server.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            name: exact name of Datasource Connection
            id: ID of Datasource Connection
        """

        if id is None and name is None:
            raise ValueError(
                "Please specify either 'id' or 'name' parameter in the constructor."
            )

        if id is None:
            objects_info = DatasourceConnection._list_datasource_connections(
                connection=connection, name=name, to_dictionary=True
            )
            if objects_info:
                object_info, object_info["connection"] = objects_info[0], connection
                self._init_variables(**object_info)
            else:
                raise ValueError(f"There is no Datasource Connection: '{name}'")
        else:
            super().__init__(connection=connection, object_id=id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.execution_mode = (
            ExecutionMode(kwargs["execution_mode"])
            if kwargs.get("execution_mode")
            else None
        )
        self.max_cancel_attempt_time = kwargs.get("max_cancel_attempt_time")
        self.max_query_exe_time = kwargs.get("max_query_exe_time")
        self.max_connection_attempt_time = kwargs.get("max_connection_attempt_time")
        self.connection_lifetime = kwargs.get("connection_lifetime")
        self.connection_idle_timeout = kwargs.get("connection_idle_timeout")
        self.char_encoding_windows = (
            CharEncoding(kwargs["char_encoding_windows"])
            if kwargs.get("char_encoding_windows")
            else None
        )
        self.char_encoding_unix = (
            CharEncoding(kwargs["char_encoding_unix"])
            if kwargs.get("char_encoding_unix")
            else None
        )
        self.table_prefix = kwargs.get("table_prefix")
        self.connection_string = kwargs.get("connection_string")
        self.parameterized_queries = kwargs.get("parameterized_queries")
        self.extended_fetch = kwargs.get("extended_fetch")
        self.driver_type = (
            DriverType(kwargs["driver_type"]) if kwargs.get("driver_type") else None
        )
        self.datasource_login = (
            DatasourceLogin.from_dict(kwargs.get("datasource_login"), self.connection)
            if kwargs.get("datasource_login")
            else None
        )
        self.database_type = kwargs.get("database_type")
        self.database_version = kwargs.get("database_version")
        self._oauth_parameter = kwargs.get("oauth_parameter")
        self.iam = kwargs.get("iam")
        self.resource = kwargs.get("resource")
        self.scope = kwargs.get("scope")
        self.enable_sso = kwargs.get("enable_sso")

    def alter(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        execution_mode: Union[str, ExecutionMode] = None,
        max_cancel_attempt_time: Optional[int] = None,
        max_query_exe_time: Optional[int] = None,
        max_connection_attempt_time: Optional[int] = None,
        connection_lifetime: Optional[int] = None,
        connection_idle_timeout: Optional[int] = None,
        char_encoding_windows: Union[str, CharEncoding] = None,
        char_encoding_unix: Union[str, CharEncoding] = None,
        table_prefix: Optional[str] = None,
        connection_string: Optional[str] = None,
        parameterized_queries: Optional[bool] = None,
        extended_fetch: Optional[bool] = None,
        driver_type: Union[str, DriverType] = None,
        database_type: Optional[str] = None,
        database_version: Optional[str] = None,
        datasource_login: Union[str, DatasourceLogin, None] = None,
        iam: Optional[dict] = None,
        resource: Optional[str] = None,
        scope: Optional[str] = None,
        enable_sso: Optional[bool] = None,
    ) -> None:
        """Alter the datasource connection properties.

        Args:
            name: Unique datasource connection name.
            description: Datasource connection description.
            execution_mode: ExecutionMode Enum specifying how SQL statements
                will be executed on a physical database connection made.
            max_cancel_attempt_time: Number of seconds before being timed out
                when attempting to make a connection.
            max_query_exe_time: Number of seconds during which a query must be
                executed before being timed out.
            max_connection_attempt_time: Number of seconds connection attempts
                to be made before timing out.
            connection_lifetime: Number of seconds of the connection lifetime.
            connection_idle_timeout: Number of seconds before being timed out
                when the connection is idle.
            char_encoding_windows: CharEncoding Enum specifying the encoding
                across connection for Windows drivers.
            char_encoding_unix: CharEncoding Enum specifying the encoding across
                connection for Unix drivers.
            table_prefix: String prefixed to the names of all temporary tables
                created by the Query Engine for this connection during report
                execution with this string.
            connection_string: Database connection string.
            parameterized_queries: Specifies whether parameterized queries are
                enabled.
            extended_fetch: Specifies whether or not to use the extended fetch
                ODBC call to retrieve data from the database connection.
            driver_type: DriverType Enum specifying Driver used for database
                connection.
            datasource_login: `DatasourceLogin` object or ID
            driver_type: ENUM Drivers used for database connection.
            database_type: Database type
            database_version: Database version
            iam: List of projects to which the fence is applied.
            resource: The url of configured Web API for OAuth authentication
                usage.
            scope: List of delegated permissions that the app is requesting.
            enable_sso: Specifies whether to use Single Sign-On.
        """
        datasource_login = (
            {'id': get_objects_id(datasource_login, DatasourceLogin)}
            if datasource_login
            else None
        )
        func = self.alter
        args = get_args_from_func(func)
        defaults = get_default_args_from_func(func)
        default_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    @classmethod
    def create(
        cls,
        connection: "Connection",
        name: str,
        description: Optional[str] = None,
        acg: Optional[int] = None,
        execution_mode: Union[str, ExecutionMode] = None,
        max_cancel_attempt_time: Optional[int] = None,
        max_query_exe_time: Optional[int] = None,
        max_connection_attempt_time: Optional[int] = None,
        connection_lifetime: Optional[int] = None,
        connection_idle_timeout: Optional[int] = None,
        char_encoding_windows: Union[str, CharEncoding] = None,
        char_encoding_unix: Union[str, CharEncoding] = None,
        table_prefix: Optional[str] = None,
        connection_string: Optional[str] = None,
        parameterized_queries: bool = False,
        extended_fetch: bool = False,
        datasource_login: Union[DatasourceLogin, str, None] = None,
        database_type: Optional[str] = None,
        database_version: Optional[str] = None,
        driver_type: Union[str, DriverType] = None,
        oauth_parameter: Optional[str] = None,
        iam: Optional[dict] = None,
        resource: Optional[str] = None,
        scope: Optional[str] = None,
        enable_sso: bool = False,
    ) -> "DatasourceConnection":
        """Create a new datasource connection on the I-Server.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
            name: Unique datasource connection name.
            description: Datasource connection description.
            acg: Object access rights. It's a bit vector where bits are defined
                at [EnumDSSXMLAccessRightFlags].
            execution_mode: ExecutionMode Enum specifying how SQL statements
                will be executed on a physical database connection made.
            max_cancel_attempt_time: Number of seconds before being timed out
                when attempting to make a connection.
            max_query_exe_time: Number of seconds during which a query must be
                executed before being timed out.
            max_connection_attempt_time: Number of seconds connection attempts
                to be made before timing out.
            connection_lifetime: Number of seconds of the connection lifetime.
            connection_idle_timeout: Number of seconds before being timed out
                when the connection is idle.
            char_encoding_windows: CharEncoding Enum specifying the encoding
                across connection for Windows drivers.
            char_encoding_unix: CharEncoding Enum specifying the encoding across
                connection for Unix drivers.
            table_prefix: String prefixed to the names of all temporary tables
                created by the Query Engine for this connection during report
                execution with this string.
            connection_string: Database connection string.
            parameterized_queries: Specifies whether parameterized queries are
                enabled.
            extended_fetch: Specifies whether or not to use the extended fetch
                ODBC call to retrieve data from the database connection.
            driver_type: DriverType Enum specifying Driver used for database
                connection.
            oauth_parameter: Used for authentication with oAuth.
            datasource_login: `DatasourceLogin` object or ID
            database_type: Database type
            database_version: Database version
            iam: List of projects to which the fence is applied.
            resource: The url of configured Web API for OAuth authentication
                usage.
            scope: List of delegated permissions that the app is requesting.
            enable_sso: Specifies whether to use Single Sign-On.

        Returns:
            DatasourceConnection object.
        """
        login_id = get_objects_id(datasource_login, DatasourceLogin)

        body = {
            "name": name,
            "description": description,
            "acg": acg,
            "executionMode": get_enum_val(execution_mode, ExecutionMode),
            "maxCancelAttemptTime": max_cancel_attempt_time,
            "maxQueryExeTime": max_query_exe_time,
            "maxConnectionAttemptTime": max_connection_attempt_time,
            "connectionLifetime": connection_lifetime,
            "connectionIdleTimeout": connection_idle_timeout,
            "charEncodingWindows": get_enum_val(char_encoding_windows, CharEncoding),
            "charEncodingUnix": get_enum_val(char_encoding_unix, CharEncoding),
            "tablePrefix": table_prefix,
            "connectionString": connection_string,
            "parameterizedQueries": parameterized_queries,
            "extendedFetch": extended_fetch,
            "database": {
                "login": {"id": login_id},
                "type": database_type,
                "version": database_version,
            },
            "driverType": get_enum_val(driver_type, DriverType),
            "oauthParameter": oauth_parameter,
            "iam": iam,
            "resource": resource,
            "scope": scope,
            "enable_sso": enable_sso,
        }
        body = helper.delete_none_values(body, recursion=True)
        response = datasources.create_datasource_connection(connection, body).json()
        if config.verbose:
            logger.info(
                f"Successfully created datasource connection named: '"
                f"{response.get('name')}' "
                f"with ID: '{response.get('id')}'"
            )
        return cls.from_dict(source=response, connection=connection)

    @method_version_handler('11.3.0100')
    def test_connection(self) -> bool:
        """Test datasource connection object.

        Returns:
            True if connection can be established, else False.
        """
        body = {"id": self.id}
        return datasources.test_datasource_connection(self.connection, body).ok

    @method_version_handler('11.3.0900')
    def convert_to_dsn_less(self):
        """Convert datasource connection from DSN to DSN-less format and update
        the object to metadata.
        """
        response = datasources.convert_connection_dsn(
            connection=self.connection, ds_connection_id=self.id
        )
        if response.ok:
            response = response.json()
            self._set_object_attributes(**response)

    @classmethod
    def _list_datasource_connections(
        cls,
        connection: "Connection",
        to_dictionary: bool = False,
        limit: Optional[int] = None,
        **filters,
    ) -> Union[List["DatasourceConnection"], List[dict]]:
        objects = helper.fetch_objects(
            connection=connection,
            api=datasources.get_datasource_connections,
            dict_unpack_value="connections",
            limit=limit,
            filters=filters,
        )
        if to_dictionary:
            return objects
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]

    @property
    def oauth_parameter(self):
        return self._oauth_parameter



"""This is the demo script to show how to manage dossiers. Its basic goal is
to present what can be done with this module and to ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.project_objects.dossier import (
    Dossier,
    list_dossiers,
    list_dossiers_across_projects
)

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project here

conn = get_connection(workstationData, PROJECT_NAME)

# Define variables which can be later used in a script
PROJECT_ID = $project_id # Insert project ID here
DOSSIER_NAME = $dossier_name  # Insert name of the Dossier here

# Dossier management
# List dossiers with different conditions
list_of_all_dossiers = list_dossiers(connection=conn)
print(list_of_all_dossiers)
list_of_all_dossiers_as_dicts = list_dossiers(connection=conn, to_dictionary=True)
print(list_of_all_dossiers_as_dicts)
list_of_all_dossiers_as_dataframes = list_dossiers(connection=conn, to_dataframe=True)
print(list_of_all_dossiers_as_dataframes)
list_of_dossiers_by_project_id = list_dossiers(connection=conn, project_id=PROJECT_ID)
print(list_of_dossiers_by_project_id)
list_of_dossiers_by_name = list_dossiers(connection=conn, name=DOSSIER_NAME)
print(list_of_dossiers_by_name)

# List of dossiers across projects, with examples of different conditions
list_of_dossiers_across_projects = list_dossiers_across_projects(connection=conn)
print(list_of_dossiers_across_projects)
list_of_dossiers_across_projects_as_dicts = list_dossiers_across_projects(
    connection=conn, to_dictionary=True
)
print(list_of_dossiers_across_projects_as_dicts)
list_of_dossiers_across_projects_as_dataframe = list_dossiers_across_projects(
    connection=conn, to_dataframe=True
)
print(list_of_dossiers_across_projects_as_dataframe)

# Define a variable which can be later used in a script
DOSSIER_ID = $dossier_id  # Insert ID of Dossier on which you want to perform actions

# Get single dossier by its id
dossier = Dossier(connection=conn, id=DOSSIER_ID)

# List dossier properties
properties = dossier.list_properties()
print(properties)

# Delete dossier without prompt
dossier.delete(force=True)

# Define variables which can be later used in a script
NEW_DOSSIER_NAME = $new_dossier_name  # Insert new name of edited dossier here
NEW_DOSSIER_DESCRIPTION = $new_dossier_description  # Insert new description of edited dossier here
FOLDER_ID = $folder_id  # Insert folder ID here

# Alter dossier
dossier.alter(name=NEW_DOSSIER_NAME)
dossier.alter(description=NEW_DOSSIER_DESCRIPTION)
dossier.alter(folder_id=FOLDER_ID)

# Define a variable which can be later used in a script
USER_ID = $user_id  # Insert user ID here

# Publish and unpublish dossier
dossier.publish()
dossier.publish(recipients=USER_ID)
dossier.unpublish()

# Share dossier with given user
dossier.share_to(users=USER_ID)

# List dossier schedules
schedules = dossier.list_available_schedules()
print(schedules)
schedules_as_dicts = dossier.list_available_schedules(to_dictionary=True)
print(schedules_as_dicts)

# List connected cubes
cubes = dossier.get_connected_cubes()
print(cubes)

# Get list of dossier cache, with examples of different conditions
list_of_all_dossier_cache = Dossier.list_caches(connection=conn)
print(list_of_all_dossier_cache)
list_of_limited_dossier_cache = Dossier.list_caches(connection=conn, limit=5)
print(list_of_limited_dossier_cache)

# Define a variable which can be later used in a script
CACHE_ID = $cache_id  # Insert ID of cache on which you want to perform actions

list_of_dossier_cache_by_id = Dossier.list_caches(connection=conn, id=CACHE_ID)
print(list_of_dossier_cache_by_id)
list_of_all_dossier_caches_as_dicts = Dossier.list_caches(connection=conn, to_dictionary=True)
print(list_of_all_dossier_caches_as_dicts)

# Helper function to list all cache status
def show_caches_status():
    caches = Dossier.list_caches(conn)
    for cache in caches:
        print(f'Cache ID: {cache.id}, {cache.status}')


# List cache properties
cache = Dossier.list_caches(connection=conn)[0]
properties = cache.list_properties()
print(properties)

# Define a variable which can be later used in a script
OTHER_CACHE_ID = $other_cache_id  # Insert ID of cache on which you want to perform actions

# Unload multiple dossier caches
Dossier.unload_caches(connection=conn, cache_ids=[CACHE_ID, OTHER_CACHE_ID])
show_caches_status()

# Define variables which can be later used in a script
CACHE_STATUS = $cache_status # Insert cache status here
USER_NAME = $user_name # Insert user name here

# Unload all dossier caches with filter examples
Dossier.unload_all_caches(connection=conn)
show_caches_status()
Dossier.unload_all_caches(connection=conn, owner=USER_NAME)
show_caches_status()
Dossier.unload_all_caches(connection=conn, status=CACHE_STATUS)
show_caches_status()

# Load multiple dossier caches
Dossier.load_caches(connection=conn, cache_ids=[CACHE_ID, OTHER_CACHE_ID])
show_caches_status()

# Load all dossier caches with filter examples
Dossier.load_all_caches(connection=conn)
show_caches_status()
Dossier.load_all_caches(connection=conn, owner=USER_NAME)
show_caches_status()
Dossier.load_all_caches(connection=conn, status=CACHE_STATUS)
show_caches_status()

# Delete multiple dossier caches
Dossier.delete_caches(connection=conn, cache_ids=[CACHE_ID, OTHER_CACHE_ID], force=True)

# Delete all dossier caches with filter examples
Dossier.delete_all_caches(connection=conn, force=True)
Dossier.delete_all_caches(connection=conn, owner=USER_NAME, force=True)
Dossier.delete_all_caches(connection=conn, status=CACHE_STATUS, force=True)



from mstrio.connection import Connection
from mstrio.utils.api_helpers import changeset_manager
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg='Error getting attribute {id} relationship.')
def get_attribute_relationships(connection: Connection, id: str, project_id: str):
    """Get relationship(s) of an attribute

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of a attribute
        project_id: Id of a project

    Return:
        HTTP response object. Expected status: 200
    """
    with changeset_manager(connection) as changeset_id:
        return connection.get(
            url=(
                f'{connection.base_url}/api/model/systemHierarchy/attributes/'
                f'{id}/relationships'
            ),
            headers={
                'X-MSTR-MS-Changeset': changeset_id,
                'X-MSTR-ProjectID': project_id,
            },
        )


@ErrorHandler(err_msg='Error updating attribute {id} relationship.')
def update_attribute_relationships(
    connection: Connection,
    id: str,
    body: dict,
):
    """Update relationship(s) of an attribute

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of a attribute
        body: JSON-formatted definition of the attribute relationships

    Return:
        HTTP response object. Expected status: 200
    """
    with changeset_manager(connection) as changeset_id:
        return connection.put(
            url=(
                f'{connection.base_url}/api/model/systemHierarchy/attributes/'
                f'{id}/relationships'
            ),
            headers={'X-MSTR-MS-Changeset': changeset_id},
            json=body,
        )



from dataclasses import dataclass
import logging
from typing import Optional

from packaging import version

from mstrio import config
from mstrio.api import subscriptions
from mstrio.connection import Connection
from mstrio.utils.entity import EntityBase, DeleteMixin
from mstrio.utils.helper import (
    delete_none_values,
    Dictable,
    exception_handler,
    fetch_objects_async,
    filter_params_for_func,
    get_valid_project_id,
)
from mstrio.utils.version_helper import class_version_handler

logger = logging.getLogger(__name__)


def list_dynamic_recipient_lists(
    connection: Connection,
    project_id: Optional[str] = None,
    project_name: Optional[str] = None,
    to_dictionary: bool = False,
    limit: Optional[int] = None,
    **filters,
) -> list["DynamicRecipientList"] | list[dict]:
    """Get list of Dynamic Recipient List objects or dicts with them.

    Note:
        When `project_id` is specified, `project_name` is omitted.
        If neither is specified then `project_id` from `connection` object
        is taken.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        project_id (str, optional): ID of the project to list the metrics from
        project_name (str, optional): name of the project
        to_dictionary (bool, optional): If True returns list of dictionaries,
            by default (False) returns DynamicRecipientList objects
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned
        **filters: Available filter parameters: ['name', 'id', 'description',
            'source_report_id', 'physical_address', 'linked_user', 'device']

    Returns:
        list with DynamicRecipientList objects or list of dictionaries
    """
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=False if project_name else True,
    )

    msg = "Error getting Dynamic Recipient List list."
    chunk_size = (
        1000
        if version.parse(connection.iserver_version) >= version.parse('11.3.0300')
        else 1000000
    )

    objects = fetch_objects_async(
        connection=connection,
        api=subscriptions.list_dynamic_recipient_lists,
        async_api=subscriptions.list_dynamic_recipient_lists_async,
        limit=limit,
        chunk_size=chunk_size,
        filters=filters,
        error_msg=msg,
        dict_unpack_value='listOfDynamicRecipientLists',
        project_id=project_id,
    )

    if to_dictionary:
        return objects
    else:
        return [
            DynamicRecipientList.from_dict(connection=connection, source=obj)
            for obj in objects
        ]


@class_version_handler('11.3.0600')
class DynamicRecipientList(EntityBase, DeleteMixin):
    """Python representation of MicroStrategy DynamicRecipientList object.

    Attributes:
        id: DynamicRecipientList's ID
        name: DynamicRecipientList's name
        description: DynamicRecipientList's description
        source_report_id: Id of the Report that is the source of the
            DynamicRecipientList
        physical_address: Physical Address for the DynamicRecipientList
        linked_user: Linked User for the DynamicRecipientList
        device: Device for the DynamicRecipientList
        recipient_name: Recipient Name for the DynamicRecipientList
        notification_address: Notification Address for the DynamicRecipientList
        notification_device: Notification Device for the DynamicRecipientList
        personalization: Personalization for the DynamicRecipientList
    """

    @dataclass
    class MappingField(Dictable):
        """Python representation of a Mapping Field.

        Attributes:
            attribute_id: ID of the mapped attribute
            attribute_form_id: ID of the mapped attribute's form
        """

        attribute_id: str
        attribute_form_id: str

    _API_GETTERS = {
        (
            'id',
            'name',
            'description',
            'source_report_id',
            'physical_address',
            'linked_user',
            'device',
            'recipient_name',
            'notification_address',
            'notification_device',
            'personalization',
        ): subscriptions.get_dynamic_recipient_list
    }
    _API_PATCH = {
        (
            'id',
            'name',
            'description',
            'source_report_id',
            'physical_address',
            'linked_user',
            'device',
            'recipient_name',
            'notification_address',
            'notification_device',
            'personalization',
        ): (subscriptions.update_dynamic_recipient_list, 'partial_put')
    }
    _API_DELETE = staticmethod(subscriptions.remove_dynamic_recipient_list)
    _FROM_DICT_MAP = {
        **EntityBase._FROM_DICT_MAP,
        'physical_address': MappingField.from_dict,
        'linked_user': MappingField.from_dict,
        'device': MappingField.from_dict,
        'recipient_name': MappingField.from_dict,
        'notification_address': MappingField.from_dict,
        'notification_device': MappingField.from_dict,
        'personalization': MappingField.from_dict,
    }

    def __init__(
        self,
        connection: Connection,
        id: Optional[str] = None,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
    ) -> None:
        """Initializes a new instance of a DynamicRecipientList class

        Args:
            connection (Connection): MicroStrategy connection object returned
                by `connection.Connection()`
            id (str, optional): DynamicRecipientList's ID. Defaults to None
            name (str, optional): DynamicRecipientList's name. Defaults to None

        Note:
            Parameter `name` is not used when fetching. If only `name` parameter
            is provided, `id` will be found automatically if such object exists.

        Raises:
            ValueError: if both `id` and `name` are not provided or if
                DynamicRecipientList with the given `name` doesn't exist.
        """
        if not id:
            if name:
                dynamic_recipient_list = self.__find_dynamic_recipient_list_by_name(
                    connection=connection, name=name
                )
                id = dynamic_recipient_list['id']
            else:
                exception_handler(
                    msg='Must provide valid id or name', exception_type=ValueError
                )
        project_id = get_valid_project_id(
            connection=connection,
            project_id=project_id,
            project_name=project_name,
            with_fallback=False if project_name else True,
        )
        super().__init__(
            connection=connection, object_id=id, name=name, project_id=project_id
        )

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.name = kwargs.get('name')
        self.project_id = kwargs.get('project_id')
        self.description = kwargs.get('description')
        self.source_report_id = kwargs.get('source_report_id')
        self.physical_address = (
            DynamicRecipientList.MappingField.from_dict(paddress)
            if (paddress := kwargs.get('physical_address'))
            else None
        )
        self.linked_user = (
            DynamicRecipientList.MappingField.from_dict(luser)
            if (luser := kwargs.get('linked_user'))
            else None
        )
        self.device = (
            DynamicRecipientList.MappingField.from_dict(dvc)
            if (dvc := kwargs.get('device'))
            else None
        )
        self.recipient_name = (
            DynamicRecipientList.MappingField.from_dict(rname)
            if (rname := kwargs.get('recipient_name'))
            else None
        )
        self.notification_address = (
            DynamicRecipientList.MappingField.from_dict(naddress)
            if (naddress := kwargs.get('notification_address'))
            else None
        )
        self.notification_device = (
            DynamicRecipientList.MappingField.from_dict(ndevice)
            if (ndevice := kwargs.get('notification_device'))
            else None
        )
        self.personalization = (
            DynamicRecipientList.MappingField.from_dict(prsnlz)
            if (prsnlz := kwargs.get('personalization'))
            else None
        )

    @classmethod
    def create(
        cls,
        connection: Connection,
        name: str,
        source_report_id: str,
        physical_address: MappingField,
        linked_user: MappingField,
        device: MappingField,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
        description: Optional[str] = None,
        recipient_name: Optional[MappingField] = None,
        notification_address: Optional[MappingField] = None,
        notification_device: Optional[MappingField] = None,
        personalization: Optional[MappingField] = None,
    ) -> "DynamicRecipientList":
        """Create a new DynamicRecipientList with specified properties.

        Args:
            name (string): DynamicRecipientList's name
            source_report_id (string): ID of the Report that is the source of
                the DynamicRecipientList
            physical_address (MappingField): Mapping Field representing the
                Physical Address for the DynamicRecipientList
            linked_user (MappingField): Mapping Field representing the Linked
                User for the DynamicRecipientList
            device (MappingField): Mapping Field representing the Device for
                the DynamicRecipientList
            project_id (string, optional): ID of the project in which the
                DynamicRecipientList is to be created, if not provided, ID will
                be taken from project_name, if neither is provided, project will
                be extracted from the connection
            project_name (string, optional): name of the project in which the
                DynamicRecipientList is to be created
            description (string, optional): DynamicRecipientList's description
            recipient_name (MappingField, optional): Mapping Field representing
                the Recipient Name for the DynamicRecipientList
            notification_address (MappingField, optional): Mapping Field
                representing the Notification Address for the
                DynamicRecipientList
            notification_device (MappingField, optional): Mapping Field
                representing the Notification Device for the
                DynamicRecipientList
            personalization (MappingField, optional): Mapping Field representing
                the Personalization for the DynamicRecipientList

        Returns:
            DynamicRecipientList class object.
        """
        body = {
            'name': name,
            'description': description if description else None,
            'sourceReportId': source_report_id,
            'physicalAddress': physical_address.to_dict(),
            'linkedUser': linked_user.to_dict(),
            'device': device.to_dict(),
            'recipientName': recipient_name.to_dict() if recipient_name else None,
            'notificationAddress': (
                notification_address.to_dict() if notification_address else None
            ),
            'notificationDevice': (
                notification_device.to_dict() if notification_device else None
            ),
            'personalization': personalization.to_dict() if personalization else None,
        }
        body = delete_none_values(source=body, recursion=True)
        project_id = get_valid_project_id(
            connection=connection,
            project_id=project_id,
            project_name=project_name,
            with_fallback=False if project_name else True,
        )
        response = subscriptions.create_dynamic_recipient_list(
            connection=connection, project_id=project_id, body=body
        ).json()

        if config.verbose:
            logger.info(
                f"Created Dynamic Recipient List named: '{name}' with ID: '"
                f"{response['id']}'"
            )

        return cls.from_dict(source={**response}, connection=connection)

    def alter(
        self,
        name: Optional[str] = None,
        source_report_id: Optional[str] = None,
        physical_address: Optional[MappingField] = None,
        linked_user: Optional[MappingField] = None,
        device: Optional[MappingField] = None,
        description: Optional[str] = None,
        recipient_name: Optional[MappingField] = None,
        notification_address: Optional[MappingField] = None,
        notification_device: Optional[MappingField] = None,
        personalization: Optional[MappingField] = None,
    ) -> None:
        """Alter a DynamicRecipientList's specified properties

        Note:
            If one alters the source_report_id, all existing Mapping Fields
            also need to be updated to reflect the new source report.

        Args:
            name (string, optional): DynamicRecipientList's name
            source_report_id (string, optional): ID of the Report that is the
                source of the DynamicRecipientList
            physical_address (MappingField, optional): Mapping Field
                representing the Physical Address for the DynamicRecipientList
            linked_user (MappingField, optional): Mapping Field representing
                the Linked User for the DynamicRecipientList
            device (MappingField, optional): Mapping Field representing the
                Device for the DynamicRecipientList
            description (string, optional): DynamicRecipientList's description
            recipient_name (MappingField, optional): Mapping Field representing
                the Recipient Name for the DynamicRecipientList
            notification_address (MappingField, optional): Mapping Field
                representing the Notification Address for the
                DynamicRecipientList
            notification_device (MappingField, optional): Mapping Field
                representing the Notification Device for the
                DynamicRecipientList
            personalization (MappingField, optional): Mapping Field
                representing the Personalization for the DynamicRecipientList
        """
        name = name or self.name
        description = description or self.description
        source_report_id = source_report_id or self.source_report_id
        physical_address = physical_address or self.physical_address
        linked_user = linked_user or self.linked_user
        device = device or self.device
        recipient_name = recipient_name or self.recipient_name
        notification_address = notification_address or self.notification_address
        notification_device = notification_device or self.notification_device
        personalization = personalization or self.personalization
        properties = filter_params_for_func(self.alter, locals(), exclude=['self'])
        self._alter_properties(**properties)

    @staticmethod
    def __find_dynamic_recipient_list_by_name(connection: "Connection", name: str):
        dynamic_recipient_lists = list_dynamic_recipient_lists(
            connection=connection, name=name
        )

        if dynamic_recipient_lists:
            number_of_drls = len(dynamic_recipient_lists)
            if number_of_drls > 1:
                raise ValueError(
                    f"There are {number_of_drls} Dynamic Recipient Lists"
                    " with this name. Please initialize with id."
                )
            else:
                return dynamic_recipient_lists[0].to_dict()
        else:
            raise ValueError(
                f"There is no DynamicRecipientList with the given name: '{name}'"
            )



"""Delete inactive caches which were not used for 30 days. IDs of deleted caches
will be printed."""

from datetime import datetime, timezone
from typing import List, Union

from mstrio.connection import Connection, get_connection
from mstrio.project_objects import CubeCache, list_cube_caches


def _get_datetime(date):
    """String to date time conversion"""
    return datetime.strptime(date, '%Y-%m-%dT%H:%M:%S.000%z')


def delete_inactive_caches(
    connection: "Connection", days_diff: str = 30, nodes: Union[str, List[str]] = None
) -> List["CubeCache"]:
    """Delete inactive caches which have `hit_count` equals 0 and their
    `last_time_updated` was earlier than `days_diff` before.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        days_diff (int, optional): number of days to determine whether to delete
            cache when its `hit_count` equals 0. Default value is 30.
        nodes (list of strings or string, optional): names of nodes from which
            caches will be deleted. By default, it equals `None` and in that
            case all nodes' names are loaded from the cluster.

    Return:
        list with cache objects which were deleted
    """
    connection._validate_project_selected()
    caches = list_cube_caches(connection, nodes)
    # delete caches which fulfill requirements to be treated as inactive
    deleted_caches = []
    for cache in caches:
        today = datetime.now(timezone.utc)
        if (
            cache.hit_count == 0
            and (today - _get_datetime(cache.last_update_time)).days > days_diff
        ):
            cache.delete(force=True)
            deleted_caches.append(cache)

    return deleted_caches


# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, 'MicroStrategy Tutorial')

# execute deletion of inactive cube caches
deleted_caches = delete_inactive_caches(conn)

print('IDs of deleted caches')
for deleted_cache in deleted_caches:
    print(deleted_cache.id)



"""This is the demo script to show how to manage contact groups.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.users_and_groups import (
    Contact, ContactGroup, ContactGroupMember, ContactGroupMemberType, list_contact_groups, User
)

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert project name here

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Define variables which can be later used in a script
MEMBER_NAME = $member_name  # here insert member name
MEMBER_ID = $member_id  # Insert member ID

member1 = ContactGroupMember(
    name=MEMBER_NAME,
    type=ContactGroupMemberType.CONTACT,  # it can be also ContactGroupMemberType.CONTACT_GROUP
    id=MEMBER_ID,
    enabled=True,
)

# Define variables which can be later used in a script
USERNAME = $username  # name of user to link to the contact group
CONTACT_GROUP_NAME = $contact_group_name  # Name of contact group
CONTACT_GROUP_DESCRIPTION = $contact_group_description  # Description of newly created contact group

# create contact group with one member represented as an object (it can be more
# and they can be also represented as dictionaries). User linked to the contact
# group can be passed as a `User` object or just its ID.

new_cg = ContactGroup.create(
    connection=conn,
    name=CONTACT_GROUP_NAME,
    linked_user=User(conn, username=USERNAME),
    members=[member1], # list of `ContactGroupMember` objects
    description=CONTACT_GROUP_DESCRIPTION,
    enabled=True,
)

# list all contact groups
contact_groups = list_contact_groups(conn)

# Define a variable which can be later used in a script
CONTACT_GROUP_ID = $contact_group_id  # Insert your contact group ID here

# There are two ways to get already existing contact group

# get contact group by ID
cg = ContactGroup(conn, id=CONTACT_GROUP_ID)

# get contact group by name
cg = ContactGroup(conn, name=CONTACT_GROUP_NAME)

# get properties of contact group
cg_name = cg.name
cg_description = cg.description
cg_lined_user = cg.linked_user
cg_members = cg.members
cg_enabled = cg.enabled

# Define variables which can be later used in a script
CONTACT_GROUP_NEW_NAME = $contact_group_new_name  # Insert new name of contact group
CONTACT_GROUP_NEW_DESCRIPTION = $contact_group_new_description  # Insert new description of contact group
OTHER_USERNAME = $other_username  # Insert new user for contact group

# alter contact group
cg.alter(
    name=CONTACT_GROUP_NEW_NAME,
    description=CONTACT_GROUP_NEW_DESCRIPTION,
    linked_user=User(conn, username=OTHER_USERNAME),
    enabled=False,
)

# Define a variable which can be later used in a script
CONTACT_NAME = $contact_name  # Insert contact name here

# more examples for contact are defined in snippet `contacts.py`

# add members to contact group - it can be `ContactGroupMember`, `Contact`
# or `ContactGroup`
contact = Contact(conn, name=CONTACT_NAME)
cg.add_members(members=[contact])  # add contact (list of `Contact` objects)
cg.add_members(members=[new_cg])  # add contact group (list of `ContactGroup` objects)

# Define a variable which can be later used in a script
OTHER_CONTACT_GROUP_NAME = $other_contact_group_name  # Insert name of another contact group here

# remove members from contact group - it can be `ContactGroupMember`, `Contact`
# or `ContactGroup`
contact_group = ContactGroup(conn, name=OTHER_CONTACT_GROUP_NAME)
cg.remove_members(members=[contact])  # remove contact
cg.remove_members(members=[contact_group])  # remove contact group

# Delete contact group. When argument `force` is set to `False` (default value),
# then deletion must be confirmed by selecting appropriate prompt value.
cg.delete(force=True)
